# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from contextlib import contextmanager

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from mvc.adapters.gtk_support.tree_view_adapters import wrap_list_property_to_treemodel
from mvc.adapters.gtk_support.dialogs.dialog_factory import DialogFactory

from pyxrd.generic.utils import not_none
from pyxrd.generic.views.treeview_tools import new_text_column, setup_treeview

from .base_controller import BaseController
from .dialog_controller import DialogController


class TreeModelMixin(object):
    """
        A mixin providing functionality to get a TreeModel property from a model.
        If that property is an actual TreeModel, it will use it directly.
        Otherwise it will first wrap it in an ObjectListStore
    """

    treemodel_getter_format = "get_%s_tree_model"

    treemodel_property_name = ""
    treemodel_class_type = None

    _treemodel = None
    @property
    def treemodel(self):
        self._update_treemodel_property()
        return self._treemodel

    @property
    def treemodel_data(self):
        if getattr(self, "model", None) is not None:
            return getattr(self.model, self.treemodel_property_name, None)
        else:
            return None

    def __init__(self, treemodel_property_name=None, treemodel_class_type=None, *args, **kwargs):
        super(TreeModelMixin, self).__init__(*args, **kwargs)
        self.treemodel_property_name = not_none(treemodel_property_name, self.treemodel_property_name)
        self.treemodel_class_type = not_none(treemodel_class_type, self.treemodel_class_type)

    def _update_treemodel_property(self):
        if getattr(self, "model", None) is not None:
            self._treemodel = wrap_list_property_to_treemodel(
                self.model,
                getattr(type(self.model), self.treemodel_property_name)
            )


class TreeViewMixin(object):
    """
        Mixin that provides some generic methods to access or set the objects selected in a treeview.
    """

    def get_selected_object(self, tv):
        # call this implementation, not the overriden method:
        objects = TreeViewMixin.get_selected_objects(self, tv)
        if objects is not None and len(objects) == 1:
            return objects[0]
        return None

    def get_selected_objects(self, tv):
        selection = tv.get_selection()
        if selection.count_selected_rows() >= 1:
            model, paths = selection.get_selected_rows()
            if hasattr(model, "get_tree_node_object_from_path"):
                return list(map(model.get_tree_node_object_from_path, paths))    
            else:
                return list(map(model.get_user_data_from_path, paths))
        return None

    def get_selected_paths(self, tv):
        selection = tv.get_selection()
        if selection.count_selected_rows() >= 1:
            model, paths = selection.get_selected_rows() # @UnusedVariable
            return paths
        return None

    def get_all_objects(self, tv):
        return tv.get_model()._data

    def set_selected_paths(self, tv, paths):
        selection = tv.get_selection()
        selection.unselect_all()
        for path in paths:
            selection.select_path(path)

class TreeControllerMixin(TreeViewMixin, TreeModelMixin):
    """
        Mixin that can be used for regular ObjectListStoreControllers (two-pane view).
        
        Attributes:
            model_property_name: the property name in the model corresponding to
             the ObjectListStore
            multi_selection: whether or not to allow multiple items to be selected
            columns: a list of tuples (name, column index or name) detailing which
             columns should be added to the TreeView. If a column name is passed,
             it is translated to the corresponding index.
             By default a text column is added, for custom setups you can define
             a custom method name according to this format: setup_treeview_col_name_%s
             Replace the %s with the column name you specified in this list.
            delete_msg: the default message to display when a user wants to delete one or more items.
            obj_type_map: a list of three-tuples (object type, view type, controller type)
             used to create the controller and view for editing a selected object.
    """

    multi_selection = True
    columns = [ ("Object name", 0) ]
    delete_msg = "Deleting objects is irreversible!\nAre You sure you want to continue?"
    obj_type_map = [] # list of three-tuples (object type, view type, controller type)

    _edit_controller = None
    _edit_view = None

    def __init__(self, *args, **kwargs):
        self.multi_selection = kwargs.pop("multi_selection", True)
        self.columns = kwargs.pop("columns", self.columns)
        self.delete_msg = kwargs.pop("delete_msg", self.delete_msg)

        super(TreeControllerMixin, self).__init__(*args, **kwargs)

    __row_signal_ids = None
    def _update_treemodel_property(self):
        # If we've connected to a treemodel before, clean up first:
        if self.__row_signal_ids is not None:
            old_treemodel, deleted_id, inserted_id = self.__row_signal_ids
            old_treemodel.disconnect(deleted_id)
            old_treemodel.disconnect(inserted_id)
            self.__row_signal_ids = None
        # If the new treemodel is set, connect it up:
        if getattr(self, "model", None) is not None:
            super(TreeControllerMixin, self)._update_treemodel_property()
            # Use private _treemodel attribute, otherwise we get infinite recursions
            self.__row_signal_ids = (
                self._treemodel,
                self._treemodel.connect("row-deleted", self.on_item_removed),
                self._treemodel.connect("row-inserted", self.on_item_inserted)
            )


    def get_new_edit_view(self, obj):
        """
            Gets a new 'edit object' view for the given obj, view and parent
            view. Default implementation loops over the `obj_type_map` attribute
            until it encounters a match.
        """
        if obj == None:
            return self.view.none_view
        else:
            for obj_tp, view_tp, ctrl_tp in self.obj_type_map: # @UnusedVariable
                if isinstance(obj, obj_tp):
                    return view_tp(parent=self.view)
            raise NotImplementedError("Unsupported object type %s; subclasses of"
                " TreeControllerMixin need to define an obj_type_map attribute!" % obj)

    def get_new_edit_controller(self, obj, view, parent=None):
        """
            Gets a new 'edit object' controller for the given obj, view and parent
            controller. Default implementation loops over the `obj_type_map` attribute
            until it encounters a match.
        """
        if obj == None:
            return None
        else:
            for obj_tp, view_tp, ctrl_tp in self.obj_type_map: # @UnusedVariable
                if isinstance(obj, obj_tp):
                    return ctrl_tp(model=obj, view=view, parent=parent)
            raise NotImplementedError("Unsupported object type; subclasses of"
                " TreeControllerMixin need to define an obj_type_map attribute!")

    def edit_object(self, obj):
        self._edit_view = self.view.set_edit_view(self.get_new_edit_view(obj))
        self._edit_controller = self.get_new_edit_controller(obj, self._edit_view, parent=self.parent)
        self._edit_view.show_all()
        return True

    def register_adapters(self):
        # connects the treeview to the treemodel
        self.setup_treeview(self.view.treeview)
        # we can now edit 'nothing':
        self.view.set_selection_state(None)
        self.edit_object(None)

    def setup_treeview(self, tv):
        """
            Sets up the treeview with columns based on the columns-tuple passed
            to the __init__ or set in the class definition.
            Subclasses can override either this method completely or provide
            custom column creation code on a per-column basis.
            To do this, create a method for e.g. column with colnr = 2:
            def setup_treeview_col_2(self, treeview, name, col_descr, col_index, tv_col_nr):
                ...
            If a string description of the column number was given, e.g. for the
            column c_name the definition should be:
            def setup_treeview_col_c_name(self, treeview, name, col_descr, col_index, tv_col_nr):
                ...
                
            The method should return True upon success or False otherwise.
        """
        sel_mode = 'MULTIPLE' if self.multi_selection else 'SINGLE' # @UndefinedVariable
        setup_treeview(
            tv, self.treemodel,
            sel_mode=sel_mode,
            on_selection_changed=self.objects_tv_selection_changed)
        tv.set_model(self.treemodel)

        # reset:
        for col in tv.get_columns():
            tv.remove_column(col)

        # add columns
        for tv_col_nr, (name, col_descr) in enumerate(self.columns):
            try:
                col_index = int(col_descr)
            except:
                col_index = getattr(self.treemodel, str(col_descr), col_descr)

            handled = False
            if hasattr(self, "setup_treeview_col_%s" % str(col_descr)):
                handler = getattr(self, "setup_treeview_col_%s" % str(col_descr))
                if callable(handler):
                    handled = handler(tv, name, col_descr, col_index, tv_col_nr)
            # custom handler failed or not present, default text column:
            if not handled:
                tv.append_column(new_text_column(
                    name, text_col=col_index,
                    resizable=(tv_col_nr == 0),
                    expand=(tv_col_nr == 0),
                    xalign=0.0 if tv_col_nr == 0 else 0.5))

        return True

    def get_selected_index(self):
        cur_obj = self.get_selected_object()
        if cur_obj is not None:
            return self.treemodel_data.index(cur_obj)
        else:
            return None

    def get_selected_object(self):
        return super(TreeControllerMixin, self).get_selected_object(self.view.treeview)

    def get_selected_objects(self):
        return super(TreeControllerMixin, self).get_selected_objects(self.view.treeview)

    def get_all_objects(self):
        return super(TreeControllerMixin, self).get_all_objects(self.view.treeview)

    def select_object(self, obj, path=None, unselect_all=True):
        selection = self.view.treeview.get_selection()
        if unselect_all: selection.unselect_all()
        if obj is not None or path is not None:
            if path is None:
                path = self.treemodel.on_get_path(obj)
            if path is not None: selection.select_path(path)

    def select_objects(self, objs):
        for obj in objs: self.select_object(obj, unselect_all=False)

    def add_object(self, new_object):
        if new_object is not None:
            index = self.get_selected_index()
            if index is not None:
                self.treemodel_data.insert(index + 1, new_object)
            else:
                self.treemodel_data.append(new_object)

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------

    def on_item_removed(self, *args):
        self.select_object(None, unselect_all=True)

    def on_item_inserted(self, model, path, iter):
        self.select_object(None, path=path, unselect_all=True)

    def objects_tv_selection_changed(self, selection):
        obj = self.get_selected_object()
        objs = self.get_selected_objects()
        self.view.set_selection_state(len(objs) if objs is not None else None)
        if self._edit_controller == None or obj != self._edit_controller.model:
            self.edit_object(obj)

    def on_load_object_clicked(self, event):
        raise NotImplementedError

    def on_save_object_clicked(self, event):
        raise NotImplementedError

    def create_new_object_proxy(self):
        raise NotImplementedError

    def on_add_object_clicked(self, event):
        new_object = self.create_new_object_proxy()
        if new_object:
            self.add_object(new_object)
            self.select_object(new_object)
        return True

    @contextmanager
    def _multi_operation_context(self):
        """
            This method should be called as a context manager (with self._multi_...)
            anytime more then one object is changed at the same time.
            Default implementation does not do anything, but this can be used
            to e.g. hold signals from firing until all objects have changed.  
        """
        yield # default implementation doesn't do anything.

    def on_del_object_clicked(self, event, del_callback=None, callback=None):
        tv = self.view.treeview
        selection = tv.get_selection()
        if selection.count_selected_rows() >= 1:
            def delete_objects(dialog):
                with self._multi_operation_context():
                    for obj in self.get_selected_objects():
                        if callable(del_callback):
                            del_callback(obj)
                        else:
                            self.treemodel_data.remove(obj)
                        if callable(callback): callback(obj)
                    self.edit_object(None)
            parent = self.view.get_top_widget()
            if not isinstance(parent, Gtk.Window): # @UndefinedVariable
                parent = None
            DialogFactory.get_confirmation_dialog(
                message=self.delete_msg, parent=parent
            ).run(delete_objects)


class ObjectListStoreController(DialogController, TreeControllerMixin):
    """
        A stand-alone, regular ObjectListStore controller (left pane with objects and right pane with object properties)
    """
    title = "Edit Dialog"
    auto_adapt = False

    def __init__(self, *args, **kwargs):
        self.title = not_none(kwargs.pop("title", None), self.title)
        super(ObjectListStoreController, self).__init__(*args, **kwargs)

    def register_view(self, view):
        super(ObjectListStoreController, self).register_view(view)
        view.set_title(self.title)

    @DialogController.model.setter
    def _set_model(self, model):
        super(ObjectListStoreController, self)._set_model(model)
        if self.view is not None:
            self._update_treemodel_property()

    def register_adapters(self):
        TreeControllerMixin.register_adapters(self)


class ChildObjectListStoreController(BaseController, TreeControllerMixin):
    """
        An embeddable, regular ObjectListStore controller (left pane with objects and right pane with object properties)
    """
    auto_adapt = False

    @DialogController.model.setter
    def _set_model(self, model):
        super(ObjectListStoreController, self)._set_model(model)
        if self.view is not None:
            self._update_treemodel_property()

    def register_adapters(self):
        TreeControllerMixin.register_adapters(self)

class InlineObjectListStoreController(BaseController, TreeControllerMixin):
    """
        ObjectListStore controller that consists of a single TreeView, 
        with import & export and add & delete buttons and an optional combo box
        for type selection
        Subclasses should override the _setup_treeview method to setup their 
        columns and edit support.
    """
    treeview = None
    enable_import = False
    enable_export = False
    add_types = list()
    auto_adapt = False

    _edit_dict = None
    def _edit_item(self, item):
        item_type = type(item)
        if self._edit_dict is None:
            # Create a edit dict which keeps track of our controllers
            self._edit_dict = {}

        # If the first time, create the view & controller
        if not item in self._edit_dict:        
            for name, tpe, view, ctrl in self.add_types: # @UnusedVariable
                if tpe == item_type:
                    vw = view()
                    ctrl = ctrl(model=item, view=vw, parent=self)
                    self._edit_dict[item] = (vw, ctrl)
                    break
        # Re-use previously created controllers
        vw, ctrl = self._edit_dict[item]
        vw.present()


    def _setup_combo_type(self, combo): # TODO this is view-related!
        if self.add_types:
            store = Gtk.ListStore(str, object, object, object) # @UndefinedVariable
            for name, type, view, ctrl in self.add_types: # @ReservedAssignment
                store.append([name, type, view, ctrl])

            combo.set_model(store)

            cell = Gtk.CellRendererText() # @UndefinedVariable
            combo.pack_start(cell, True)
            combo.add_attribute(cell, 'text', 0)

            def on_changed(combo, user_data=None):
                itr = combo.get_active_iter()
                if itr is not None:
                    val = combo.get_model().get_value(itr, 1)
                    self.add_type = val
            combo.connect('changed', on_changed)
            combo.set_active_iter(store[0].iter)
            combo.set_visible(True)
            combo.set_no_show_all(False)
            combo.show_all()

    def _setup_treeview(self, tv, model):
        raise NotImplementedError

    def __init__(self, *args, **kwargs):
        self.enable_import = kwargs.pop("enable_import", self.enable_import)
        self.enable_export = kwargs.pop("enable_export", self.enable_export)
        super(InlineObjectListStoreController, self).__init__(*args, **kwargs)

    @BaseController.model.setter
    def _set_model(self, model):
        super(ObjectListStoreController, self)._set_model(model)
        if self.view is not None:
            self._update_treemodel_property()

    def register_adapters(self):
        if self.treemodel is not None:
            self.treeview = self.view.treeview
            self.treeview.connect('cursor-changed', self.on_treeview_cursor_changed, self.treemodel)
            self._setup_treeview(self.treeview, self.treemodel)
            self.type_combobox = self.view.type_combobox_widget
            self._setup_combo_type(self.type_combobox)
            self.update_sensitivities()
        return

    def update_sensitivities(self):
        self.view.del_item_widget.set_sensitive((self.treeview.get_cursor() != (None, None)))
        self.view.add_item_widget.set_sensitive((self.treemodel is not None))
        self.view.export_items_widget.set_visible(self.enable_export)
        self.view.export_items_widget.set_sensitive(len(self.treemodel_data) > 0)
        self.view.import_items_widget.set_visible(self.enable_import)

    def get_selected_object(self):
        return TreeViewMixin.get_selected_object(self, self.treeview)

    def get_selected_objects(self):
        return TreeViewMixin.get_selected_objects(self, self.treeview)

    def get_all_objects(self):
        return TreeViewMixin.get_all_objects(self, self.treeview)

    def select_object(self, obj, path=None, unselect_all=True):
        selection = self.treeview.get_selection()
        if unselect_all: selection.unselect_all()
        if obj is not None and hasattr(self.treemodel, "on_get_path"):
            selection.select_path(self.treemodel.on_get_path(obj))
        elif path is not None:
            selection.select_path(path)

    def create_new_object_proxy(self):
        raise NotImplementedError

    def edit_object(self, obj):
        pass

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_treeview_cursor_changed(self, widget, model):
        self.update_sensitivities()

    def on_item_cell_edited(self, cell, path, new_text, model, col):
        model.set_value(model.get_iter(path), col, model.convert(col, new_text))
        pass

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

from .refine_method_option import RefineMethodOption

class RefineMethodMeta(type):
    """
        The metaclass for creating a RefineMethod (sub)class
        Will register the class type so we can build a list of RefineMethod 
        classes dynamically.
        If the (sub)class does not want to be registered, it should set
        the 'disabled' class attribute to True.
    """
    registered_methods = {}

    def __new__(meta, name, bases, class_dict):  # @NoSelf
        options = []
        for name, value in class_dict.items():
            if isinstance(value, RefineMethodOption):
                options.append(name)
                setattr(value, 'label', name)
        class_dict['options'] = options

        cls = type.__new__(meta, name, bases, class_dict)

        if not getattr(cls, 'disabled', False):
            meta.registered_methods[getattr(cls, 'index')] = cls

        return cls

    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .xrd_data_object import XRDDataObject

class XRDParserMixin(object):
    """
        This is a mixin class which provides common functionality and attributes
        for XRD-data parser classes.
    """

    data_object_type = XRDDataObject

    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#  Copyright (C) 2005 by Roberto Cavada <roboogle@gmail.com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import logging
from mvc.support.gui_loop import add_idle_call
logger = logging.getLogger(__name__)

from .observers import Observer
from .support.exceptions import TooManyCandidatesError

class Controller(Observer):

    auto_adapt = True
    auto_adapt_included = None
    auto_adapt_excluded = None

    register_lazy = True

    __adapters = None
    __parsed_user_props = None
    _controller_scope_aplied = False

    @property
    def __user_props(self):
        assert not (self.auto_adapt_included is not None and self.auto_adapt_excluded is not None), \
            "Controller '%s' has set both auto_adapt_included and auto_adapt_excluded!" % self
        assert self.model is not None, \
            "Controller '%s' has None as model! Did you forget to pass it as a keyword argument?" % self
        if not self._controller_scope_aplied:
            props = [prop.label for prop in self.model.Meta.all_properties]
            if self.auto_adapt_included is not None:
                self.__parsed_user_props = self.__parsed_user_props.union(set([p for p in props if p not in self.auto_adapt_included]))
            elif self.auto_adapt_excluded is not None:
                self.__parsed_user_props = self.__parsed_user_props.union(set([p for p in props if p in self.auto_adapt_excluded]))
            self._controller_scope_aplied = True
        return self.__parsed_user_props

    @__user_props.setter
    def __user_props(self, value):
        return # ignore
        self.__parsed_user_props = set(value)


    _model = None
    def _get_model(self):
        return self._model
    def _set_model(self, model):
        if self._model is not None:
            self._clear_adapters()
            self.relieve_model(self._model)
        self._model = model
        if self._model is not None:
            self.observe_model(self._model)
            if self.view is not None:
                self.register_adapters()
                if self.auto_adapt: self.adapt()
    def _del_model(self):
        del self._model
        self._model = None
    model = property(_get_model, _set_model, _del_model)

    __view = None
    def _get_view(self):
        return self.__view
    def _set_view(self, view):
        if self.__view != view:
            self.__view = view
            if self.__view is not None:
                if self.register_lazy:
                    add_idle_call(self._idle_register_view, self.__view)
                else:
                    self._idle_register_view(self.__view)
    def _del_view(self):
        self.__view = None
    view = property(_get_view, _set_view, _del_view, "This controller's view")

    def __init__(self, *args, **kwargs):
        """
        Two positional and two optional keyword arguments.
        
        *model* will have the new instance registered as an observer.
        It is made available as an attribute.
        
        *view* may contain signal connections loaded from XML. The handler
        methods have to exist in this class.
        
        *spurious* denotes whether notifications in this class will be called
        if a property of *model* is set to the same value it already has.
        
        *auto_adapt* denotes whether to call :meth:`adapt` with no arguments
        as part of the view registration process.

        View registration consists of connecting signal handlers,
        :meth:`register_view` and :meth:`register_adapters`, and is scheduled
        with the GTK main loop. It happens as soon as possible but after the
        constructor returns. When it starts *view* is available as an
        attribute.
        
        Subclasses can also override either the `auto_adapt_included` or
        `auto_adapt_excluded` class attributes. These should be either None or
        contain a list of properties to exclude (or include) from auto
        adaptation for this controller. If not specified, all properties are
        adapted if auto_adapt is True.
        """

        # set of properties explicitly adapted by the user:
        self.__adapters = []
        self.__parsed_user_props = set()

        # Some general keyword arguments:
        self.parent = kwargs.pop("parent", None)
        self.auto_adapt = kwargs.pop("auto_adapt", self.auto_adapt)
        self.register_lazy = kwargs.pop("register_lazy", self.register_lazy)

        # Pop the view, keep the model as we pass it down to the observer
        _view = kwargs.pop("view", None)
        _model = kwargs.get("model", None)

        # Init base classes
        super(Controller, self).__init__(*args, **kwargs)

        # Set model (will register controller as observer)
        self.model = _model
        # Set view (will register & adapt to view)
        self.view = _view

        return

    def _idle_register_view(self, view):
        """Internal method that calls register_view"""
        assert(self.view is not None)
        if self.model is not None:
            self.__autoconnect_signals()

            self.register_view(self.view)
            self.register_adapters()
            if self.auto_adapt: self.adapt()
            return False
        else:
            return True

    def _clear_adapters(self):
        """Clears & disconnects all adapters from this controller"""
        if self.__adapters:
            for ad in self.__adapters:
                ad.disconnect()
            self.__adapters[:] = []

    def register_view(self, view):
        """
        This does nothing. Subclasses can override it to connect signals
        manually or modify widgets loaded from XML, like adding columns to a
        TreeView. No super call necessary.
        
        *view* is a shortcut for ``self.view``.
        """
        assert(self.model is not None)
        assert(self.view is not None)
        return

    def register_adapters(self):
        """
        This does nothing. Subclasses can override it to create adapters.
        No super call necessary.
        """
        assert(self.model is not None)
        assert(self.view is not None)
        return

    def adapt(self, *args):
        """
        There are four ways to call this:

        .. method:: adapt()
           :noindex:

           Take properties from the model for which ``adapt`` has not yet been
           called, match them to the view by name, and create adapters fitting
           for the respective widget type.
           
           That information comes from :mod:`mvc.adapters.default`.
           See :meth:`_find_widget_match` for name patterns.

           .. versionchanged:: 1.99.1
              Allow incomplete auto-adaption, meaning properties for which no
              widget is found.

        .. method:: adapt(ad)
           :noindex:
        
           Keep track of manually created adapters for future ``adapt()``
           calls.
        
           *ad* is an adapter instance already connected to a widget.

        .. method:: adapt(prop_name)
           :noindex:

           Like ``adapt()`` for a single property.

           *prop_name* is a string.

        .. method:: adapt(prop_name, wid_name)
           :noindex:

           Like ``adapt(prop_name)`` but without widget name matching.
           
           *wid_name* has to exist in the view.
        """

        # checks arguments
        n = len(args)
        if n not in list(range(3)): raise TypeError("adapt() takes 0, 1 or 2 arguments (%d given)" % n)

        if n == 0:
            adapters = []
            props = self.model.Meta.get_viewable_properties()
            # matches all properties not previously adapted by the user:
            for prop in [p for p in props if p.label not in self.__user_props]:
                try: wid_name = self._find_widget_match(prop)
                except TooManyCandidatesError as e:
                    # multiple candidates, gives up
                    raise e
                except ValueError as e:
                    # no widgets found for given property, continue after emitting a warning
                    if e.args:
                        logger.warn(e[0])
                    else:
                        logger.warn("No widget candidates match property '%s'"
                            % prop.label)
                else:
                    logger.debug("Auto-adapting property %s and widget %s" % \
                                     (prop.label, wid_name))
                    adapters += self.__create_adapters__(prop, wid_name)
                    pass
                pass

        elif n == 1: # one argument
            from .adapters import AbstractAdapter

            if isinstance(args[0], AbstractAdapter): adapters = (args[0],)

            elif isinstance(args[0], str):
                prop = getattr(type(self.model), args[0])
                wid_name = self._find_widget_match(prop)
                adapters = self.__create_adapters__(prop, wid_name)
                pass
            else: raise TypeError("Argument of adapt() must be either an Adapter or a string")

        else: # two arguments
            if not (isinstance(args[0], str) and isinstance(args[1], str)):
                raise TypeError("Arguments of adapt() must be two strings")

            # retrieves both property and widget, and creates an adapter
            prop_name, wid_name = args #@UnusedVariable
            adapters = self.__create_adapters__(prop, wid_name)
            pass

        for ad in adapters:
            self.__adapters.append(ad)
            # remember properties added by the user
            if n > 0: self.__user_props.add(ad.get_property_name())
            pass

        return

    def _find_widget_match(self, prop):
        """
        Checks if the view has defined a 'widget_format' attribute (e.g. 
        "view_%s") If so, it uses this format to search for the widget in 
        the view, if not it takes the *first* widget with a name ending with the
        property name.
        """

        widget_name = None
        widget_format = getattr(self.view, 'widget_format', "%s")

        if widget_format:
            widget_name = widget_format % prop.label
            widget = self.view[widget_name]
            if widget is None: # not in view
                if prop is not None and prop.widget_type == 'scale':
                    self.view.add_scale_widget(prop)
                else:
                    widget_name = None

        else:
            for wid_name in self.view:
                if wid_name.lower().endswith(prop.label.lower()):
                    widget_name = wid_name
                    break

        if widget_name == None:
            logger.setLevel(logging.INFO)
            raise ValueError("No widget candidates match property '%s'" % prop.label)

        return widget_name

    # performs Controller's signals auto-connection:
    def __autoconnect_signals(self):
        """This is called during view registration, to autoconnect
        signals in glade file with methods within the controller"""
        dic = {}
        for name in dir(self):
            method = getattr(self, name)
            if (not callable(method)): continue
            assert(name not in dic) # not already connected!
            dic[name] = method
            pass

        # Auto connect builder if available:
        if self.view._builder is not None:
            self.view._builder.connect_signals(dic)
            pass

        return

    def _get_handler_list(self):
        from .adapters import AdapterRegistry

        # Add default widget handlers:
        local_handlers = {}
        adapter_registry = AdapterRegistry.get_selected_adapter_registry()
        local_handlers.update(adapter_registry)

        # Override with class instance widget handlers:
        for widget_type, handler in self.widget_handlers.items():
            if isinstance(handler, str):
                self.widget_handlers[widget_type] = getattr(self, handler)
        local_handlers.update(self.widget_handlers)

        return local_handlers

    def __create_adapters__(self, prop, wid_name):
        """
            Private service that looks at property and widgets types,
            and possibly creates one or more (best) fitting adapters
            that are returned as a list.
        """
        try:
            logger.debug("Adapting property %s to widget names '%s'" % (prop.label, wid_name))
            if prop.visible:

                wid = self.view[wid_name]
                if wid == None:
                    raise ValueError("Widget '%s' not found in view '%s' by controller '%s'" % (wid_name, self.view, self))

                local_handlers = self._get_handler_list()

                handler = local_handlers.get(prop.widget_type)
                ad = handler(self, prop, wid)

                return [ad]
            else:
                return []
        except BaseException as error:
            raise RuntimeError("Unhandled error in '%s'.__create_adapters__ for property '%s' and widget '%s'!" % (type(self), prop.label, wid_name)) from error

    pass # end of class Controller



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#  Copyright (C) 2005 by Tobias Weber
#  Copyright (C) 2005 by Roberto Cavada <roboogle@gmail.com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import weakref

class ObsWrapperBase (object):
    """
    This class is a base class wrapper for user-defined classes and
    containers like lists, maps, signals, etc.
    """

    def __init__(self):
        # all model instances owning self (can be multiple due to
        # inheritance). Each element of the set is a pair (model,
        # property-name)
        self.__models = set()
        return

    def __add_model__(self, model, prop_name):
        """Registers the given model to hold the wrapper among its
        properties, within a property whose name is given as well"""
        ref = weakref.ref(model)
        self.__models.add((ref, prop_name))
        return

    def __remove_model__(self, model, prop_name):
        """Unregisters the given model, to release the wrapper. This
        method reverts the effect of __add_model__"""
        delete_models = set()
        for ref, prop_name in self.__models:
            if ref() == model and prop_name == prop_name:
                delete_models.add((ref, prop_name))
        self.__models -= delete_models
        return

    def __get_models__(self):
        delete_models = set()
        for ref, prop_name in self.__models:
            model = ref()
            if model is not None:
                yield model, prop_name
            else:
                delete_models.add((ref, prop_name))
        self.__models -= delete_models

    def _notify_method_before(self, instance, name, args, kwargs):
        for m, n in self.__get_models__():
            m.notify_method_before_change(n, instance, name,
                                          args, kwargs)
            pass
        return

    def _notify_method_after(self, instance, name, res_val, args, kwargs):
        for m, n in self.__get_models__():
            m.notify_method_after_change(n, instance, name, res_val,
                                         args, kwargs)
            pass
        return

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from io import StringIO

from ..data_object import DataObject
from pyxrd.generic.utils import not_none

class XRDDataObject(DataObject):
    """
        A generic class holding all the information retrieved from an XRD data
        file using an XRD-parser class.
    """

    # General information
    name = None
    date = None

    # x-axis range
    twotheta_min = None
    twotheta_max = None
    twotheta_count = None
    twotheta_step = None

    # Wavelength and target information
    target_type = None
    alpha1 = None
    alpha2 = None
    alpha_average = None
    beta = None
    alpha_factor = None

    # Goniometer setup
    radius = None
    soller1 = None
    soller2 = None
    divergence = None

    # Generator or list of x,y data
    data = None

    def create_gon_file(self):

        output = """        {
            "type": "Goniometer", 
            "properties": {
                "radius": %(radius)f, 
                "divergence": %(divergence)f, 
                "soller1": %(soller1)f, 
                "soller2": %(soller2)f, 
                "min_2theta": %(twotheta_min)f, 
                "max_2theta": %(twotheta_max)f, 
                "steps": %(twotheta_count)f, 
                "wavelength": %(alpha_average)f, 
                "has_ads": false, 
                "ads_fact": 1.0, 
                "ads_phase_fact": 1.0, 
                "ads_phase_shift": 0.0, 
                "ads_const": 0.0
            }
        }""" % dict(
            radius=float(not_none(self.radius, 25)),
            divergence=float(not_none(self.divergence, 0.5)),
            soller1=float(not_none(self.soller1, 2.5)),
            soller2=float(not_none(self.soller2, 2.5)),
            twotheta_min=float(not_none(self.twotheta_min, 3.0)),
            twotheta_max=float(not_none(self.twotheta_max, 45.0)),
            twotheta_count=float(not_none(self.twotheta_count, 2500)),
            alpha_average=float(not_none(self.alpha_average, 0.154056)),
        )
        f = StringIO(output)
        f.flush()
        return f


    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk, Gdk, GdkPixbuf, GObject # @UnresolvedImport

from time import time, sleep
from pyxrd.application.icons import get_icon_list

def scale_ratio(src_width, src_height, dest_width, dest_height):
    """Return a size fitting into dest preserving src's aspect ratio."""
    if src_height > dest_height:
        if src_width > dest_width:
            ratio = min(float(dest_width) / src_width,
                        float(dest_height) / src_height)
        else:
            ratio = float(dest_height) / src_height
    elif src_width > dest_width:
        ratio = float(dest_width) / src_width
    else:
        ratio = 1
    return int(ratio * src_width), int(ratio * src_height)

class ScalableImage(Gtk.Image):
    """A Gtk.Image that rescales to fit whatever size is available.

    Only Pixbuf data is supported; it can be loaded from a file or
    passed directly.
    """
    def __init__(self, pixbuf=None):
        super(ScalableImage, self).__init__()
        self._pixbuf = None
        self.connect('size-allocate', self._on_size_allocate)
        self.set_size_request(1, 1)
        self._hyper_id = None
        if pixbuf is not None:
            self.set_from_pixbuf(pixbuf)

    def _on_timeout_hyper(self):
        """Perform a delayed high-quality scale."""
        self._hyper_id = None
        allocation = self.get_allocation()
        target_width, target_height = scale_ratio(
            self._pixbuf.get_width(), self._pixbuf.get_height(),
            allocation.width, allocation.height)
        if target_width > 0 and target_height > 0:
            pixbuf = self._pixbuf.scale_simple(
                target_width, target_height, GdkPixbuf.InterpType.HYPER) # @UndefinedVariable
            super(ScalableImage, self).set_from_pixbuf(pixbuf)

    def _on_size_allocate(self, image, allocation, force=False):
        """Scale the internal pixbuf copy to a new size."""
        if self._pixbuf is None:
            return
        pix_width = self._pixbuf.get_width()
        pix_height = self._pixbuf.get_height()
        target_width, target_height = scale_ratio(
            pix_width, pix_height, allocation.width, allocation.height)
        old_pix = self.get_pixbuf()
        if target_width < pix_width or target_height < pix_height:
            if (force or not old_pix
                or old_pix.get_width() != target_width
                or old_pix.get_height() != target_height):
                # If we're forcing an update we have a new image,
                # since that's a "big" event we can afford to
                # hyper-scale right away. On the other hand if it's
                # not a forced update the window just resized, that
                # needs responsiveness immediately, so delay the hyper
                # scale until the window is stationary for at least
                # 1/10 of a second.
                if self._hyper_id:
                    GObject.source_remove(self._hyper_id)
                    self._hyper_id = None
                if target_width > 0 and target_height > 0:
                    pixbuf = self._pixbuf.scale_simple(
                        target_width, target_height,
                        GdkPixbuf.InterpType.HYPER if force else GdkPixbuf.InterpType.NEAREST) # @UndefinedVariable
                    if not force:
                        self._hyper_id = GObject.timeout_add(100, self._on_timeout_hyper)
                    super(ScalableImage, self).set_from_pixbuf(pixbuf)
        elif old_pix != self._pixbuf:
            if self._hyper_id:
                GObject.source_remove(self._hyper_id)
                self._hyper_id = None
            super(ScalableImage, self).set_from_pixbuf(self._pixbuf)

    def set_from_file(self, filename):
        """Set the image by loading a file."""
        pixbuf = GdkPixbuf.Pixbuf.new_from_file(filename) #@UndefinedVariable
        self.set_from_pixbuf(pixbuf)

    def set_from_pixbuf(self, pixbuf):
        """Set the image from a GdkPixbuf.Pixbuf."""
        self._pixbuf = pixbuf
        self._on_size_allocate(None, self.get_allocation(), force=True)

    def __not_implemented(self, *args):
        """This Gtk.Image storage type is not supported."""
        raise NotImplementedError("only pixbuf images are currently supported")

    set_from_animation = __not_implemented
    set_from_gicon = __not_implemented
    set_from_icon_name = __not_implemented
    set_from_icon_set = __not_implemented
    set_from_image = __not_implemented
    set_from_pixmap = __not_implemented
    set_from_stock = __not_implemented

class SplashScreen(object):
    def __init__(self, filename, version=""):
        # DONT connect 'destroy' event here!
        self.window = Gtk.Window(Gtk.WindowType.TOPLEVEL)
        self.window.set_auto_startup_notification(False)
        self.window.set_default_icon_list(get_icon_list())
        self.window.set_icon_list(get_icon_list())
        self.window.set_title('PyXRD')
        self.window.set_skip_taskbar_hint(True)
        self.window.set_position(Gtk.WindowPosition.CENTER)
        self.window.set_decorated(False)
        self.window.set_resizable(False)
        self.window.set_border_width(1)
        self.window.modify_bg(Gtk.StateType.NORMAL, Gdk.color_parse('white')) # @UndefinedVariable

        ebox = Gtk.EventBox() # prevent the black color from showing through...
        self.window.add(ebox)

        main_vbox = Gtk.VBox(False, 1)
        main_vbox.set_border_width(10)
        ebox.add(main_vbox)

        self.img = ScalableImage()
        self.img.set_from_file(filename)
        self.img.set_size_request(500, 300)
        main_vbox.pack_start(self.img, True, True, 0)

        self.lbl = Gtk.Label()
        self.lbl.set_markup("<span size=\"larger\"><b>Loading ...</b></span>")
        self.lbl.set_alignment(0.5, 0.5)
        main_vbox.pack_end(self.lbl, True, True, 0)

        self.version_lbl = Gtk.Label()
        self.version_lbl.set_markup("<i>Version %s</i>" % version)
        self.version_lbl.set_alignment(0.5, 0.5)
        main_vbox.pack_end(self.version_lbl, True, True, 0)

        self.window.show_all()
        while Gtk.events_pending():
            Gtk.main_iteration()
        self.start_time = time()

        self.closed = False
    
    def set_message(self, message):
        self.lbl.set_markup("<span size=\"larger\"><b>%s</b></span>" % message)
        while Gtk.events_pending():
            Gtk.main_iteration()

    def close(self):
        if not self.closed:
            self.window.set_auto_startup_notification(True)
            while (max(5 - (time() - self.start_time), 0) != 0):
                sleep(0.1)
                while Gtk.events_pending():
                    Gtk.main_iteration()
            self.window.destroy()
            del self.window, self.lbl, self.img, self.version_lbl
            self.closed = True


    pass # end of class



#    This file is part of DEAP.
#
#    DEAP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    DEAP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with DEAP. If not, see <http://www.gnu.org/licenses/>.

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

import numpy as np
import scipy
import itertools
import random

from pyxrd.refinement.refine_method import RefineMethod
from pyxrd.refinement.refine_method_option import RefineMethodOption
from pyxrd.refinement.refine_async_helper import RefineAsyncHelper

from deap import base, creator, tools #@UnresolvedImport

from .deap_utils import pyxrd_array, PyXRDParetoFront, FitnessMin, result_func

# Default settings:
NGEN = 100
NSWARMS = 1
NEXCESS = 3
NPARTICLES = 15
CONV_FACTR = 0.3 # 5.*1e-2

class MultiPSOStrategy(object):

    def generate_particle(self, pclass, dim, pmin, pmax, smin, smax):
        """ Generate a particle """
        part = pclass(random.uniform(pmin[i], pmax[i]) for i in range(dim))
        part.speed = np.array([random.uniform(smin[i], smax[i]) for i in range(dim)])
        return part

    def update_particle(self, part, best, chi, c):
        """
            Update a particle's position & speed
            part: the particle
            best: the global best
            chi ~ recombination factor
            c ~ scale factor(s)
        """
        part_pos = np.asarray(part)
        best_pos = np.asarray(best)
        pers_pos = np.asarray(part.best)
        speed = np.asarray(part.speed)

        ce1 = c * np.random.uniform(0, 1, size=len(part))
        ce2 = c * np.random.uniform(0, 1, size=len(part))
        ce1_p = ce1 * np.array(best_pos - part_pos)
        ce1_g = ce2 * np.array(pers_pos - part_pos)

        f = (ce1_p + ce1_g)

        # Calculate velocity:
        a = chi * (f + speed) - speed

        # Adjust speed:
        speed = a + speed
        # Set position & speed:
        part[:] = part_pos + speed
        part.speed = speed
        # Clear fitness:
        del part.fitness.values

    def create_swarm(self, container, iterable):
        """ Returns a swarm container using the iterable """
        return container(iterable)

    def generate_particles(self, func, n):
        """ Returns a particle generator """
        for _ in range(n):
            yield func()

    def update_swarm(self, swarm, part):
        """ Update swarm's attractors personal best and global best """
        if not part.fitness.valid:
            raise RuntimeError("Particles need to have a valid fitness before calling update_swarm!")
        if part.best == None or part.fitness > part.bestfit:
            part.best = creator.Particle(part)          # Get the position @UndefinedVariable
            part.bestfit.values = part.fitness.values   # Get the fitness
        if swarm.best == None or part.fitness > swarm.bestfit:
            swarm.best = creator.Particle(part)         # Get the position @UndefinedVariable
            swarm.bestfit.values = part.fitness.values  # Get the fitness

    def give_reinit_swarms(self, population):
        """ Gives a set of swarm indeces that need to be reinitialized (overlap)"""
        reinit_swarms = set()
        for s1, s2 in itertools.combinations(list(range(len(population))), 2):
            # Swarms must have a best and not already be set to reinitialize
            if population[s1].best is not None and population[s2].best is not None and not (s1 in reinit_swarms or s2 in reinit_swarms):
                # if t-test is True, then we reinit the worst of the two swarms
                t, _ = scipy.stats.ttest_ind(population[s1], population[s2]) #@UndefinedVariable
                if np.all(t < 0.1):
                    if population[s1].bestfit <= population[s2].bestfit:
                        reinit_swarms.add(s1)
                    else:
                        reinit_swarms.add(s2)
        return reinit_swarms

    def give_converged_swarms(self, population, conv_factr, converged_bests=[]):
        """ Returns the number of converged swarms and the worst swarm index """
        # Convergence check:
        not_converged = 0
        worst_swarm_idx = None
        worst_swarm = None
        for i, swarm in enumerate(population):
            # Compute the diameter of the swarm:
            std = np.std([ind.fitness.values for ind in swarm])
            # If it's larger then a given factor, we've not converged yet:
            if std > conv_factr:
                not_converged += 1
                if worst_swarm is None or swarm.bestfit < worst_swarm.bestfit:
                    worst_swarm_idx = i
                    worst_swarm = swarm
            else:
                converged_bests.append(swarm.best)

        converged_bests.sort(key=lambda i: i.fitness)

        return not_converged, worst_swarm_idx, converged_bests

    pass #end of class

class MPSOAlgorithm(RefineAsyncHelper):
    """
        Multi-particle-swarm optimization method adapted from the examples found
        in the DEAP project. Employs a T-test (two independent sample lists) to
        differentiate between swarms instead of the diameter and average of the
        swarms. Seemed to work better for scaled parameters (YMMV).
        
        Implementation of the Multiswarm Particle Swarm Optimization algorithm as
        presented in *Blackwell, Branke, and Li, 2008, Particle Swarms for Dynamic
        Optimization Problems.*
    """

    gen = -1
    converged_bests = None

    #--------------------------------------------------------------------------
    #    Initialization
    #--------------------------------------------------------------------------
    def __init__(self, toolbox, bounds, norms,
        ngen=NGEN, nswarms=NSWARMS, nexcess=NEXCESS,
        nparticles=NPARTICLES, conv_factr=CONV_FACTR,
        stats=None, halloffame=None, verbose=True, refiner=None, stop=None):
        """
            TODO
        """
        self.converged_bests = []
        self.toolbox = toolbox
        self.bounds = bounds
        self.norms = norms
        self.ngen = ngen
        self.nswarms = nswarms
        self.nexcess = nexcess
        self.nparticles = nparticles
        self.conv_factr = conv_factr
        self.stats = stats
        self.halloffame = halloffame
        self.verbose = verbose
        self.refiner = refiner

        self._stop = stop

    #--------------------------------------------------------------------------
    #    Convenience functions:
    #--------------------------------------------------------------------------

    def _evaluate_swarms(self, population):
        # Only evaluate invalid particles:
        def give_unevaluated_particles():
            for p in itertools.chain(*population):
                if p.fitness.valid: continue
                else: yield p
        self.do_async_evaluation(iter_func=give_unevaluated_particles, result_func=result_func)
        for swarm in population:
            for part in swarm:
                self.toolbox.update_swarm(swarm, part)
        return population

    def _create_and_evaluate_swarm(self):
        """ Helper function that creates, evaluates and returns a new swarm """
        particles = self.do_async_evaluation(iter_func=self.toolbox.generate_particles, result_func=result_func)
        return self.toolbox.swarm(particles)

    def _create_and_evaluate_population(self):
        """ Helper function that creates, evaluates and returns a population of swarms """
        population = [self._create_and_evaluate_swarm() for _ in range(self.nswarms)]
        if self.halloffame is not None:
            self.halloffame.update(itertools.chain(*population))
        return population

    #--------------------------------------------------------------------------
    #    Run method:
    #-------------------------------------------------------------------------
    def run(self):
        """Will run this algorithm"""
        self._setup_logging()
        population = []
        for _ in range(self.ngen):

            # Check if the user has cancelled:
            if self._user_cancelled():
                logger.info("User cancelled execution of PCMA-ES, stopping ...")
                break

            #ASK: Generate a new population:
            population = self._ask(population)
            #RECORD: For stagnation checking & logging:
            self._record(population)
            #CHECK: whether we are stagnating:
            if self._is_stagnating():
                break
            #TELL: Update the strategy with the evaluated individuals
            self._tell(population)

        return (
            self.refiner.history.best_solution,
            list(itertools.chain(*population)),
            self.converged_bests
        )

    #--------------------------------------------------------------------------
    #    Ask, tell & record:
    #--------------------------------------------------------------------------
    def _ask(self, population):
        """
            Calculates how many swarms have converged, and keeps track of the
            worst swarm. If all swarms have converged, it will add a new swarm.
            If too many swarms are roaming, it will remove the worst.
        """

        self.gen += 1
        self.refiner.status.message = "Creating generation #%d" % (self.gen + 1)

        if not population:
            # First iteration: create a new population of swarms
            population = self._create_and_evaluate_population()
        else:
            # Second and later iterations: check for overlapping swarms
            reinit_swarms = self.toolbox.give_reinit_swarms(population)
            # Reinitialize and evaluate swarms
            for sindex in reinit_swarms:
                population[sindex] = self._create_and_evaluate_swarm()

        # Get unconverged swarm count and worst swarm id:
        not_converged, worst_swarm_idx, self.converged_bests = self.toolbox.give_converged_swarms(population, self.conv_factr, self.converged_bests)

        # If all swarms have converged, add a swarm:
        if not_converged == 0:
            population.append(self._create_and_evaluate_swarm())

        # If too many swarms are roaming, remove the worst swarm:
        elif not_converged > self.nexcess:
            population.pop(worst_swarm_idx)

        return population

    def _tell(self, population):
        # Update and evaluate the swarm
        for swarm in population:
            # Update particles and swarm:
            for part in swarm:
                if swarm.best is not None and part.best is not None:
                    self.toolbox.update_particle(part, swarm.best)
        self._evaluate_swarms(population)
        self.halloffame.update(itertools.chain(*population))

    def _is_stagnating(self):
        return False # TODO FIXME

    def _setup_logging(self):
        if self.verbose:
            column_names = ["gen", "nswarm", "indiv"]
            if self.stats is not None:
                column_names += list(self.stats.functions.keys())
            self.logbook = tools.Logbook()
            self.logbook.header = column_names

    def _record(self, population):
        # Get pop size:
        pop_size = len(population)

        # Get the best solution so far:
        best = self.halloffame.get_best()
        best_f = best.fitness.values[0]

        # Calculate stats & print something if needed:
        pop = list(itertools.chain(*population))
        record = self.stats.compile(pop)
        if self.verbose:
            self.logbook.record(gen=self.gen, nswarm=pop_size, indiv=len(pop), **record)
            print(self.logbook.stream)

        # Update the context:
        self.refiner.update(best, iteration=self.gen, residual=best_f)

    pass #end of class


class RefineMPSORun(RefineMethod):
    """
        The DEAP MPSO algorithm implementation
    """
    name = "MPSO refinement"
    description = "This algorithm uses the MPSO refinement strategy"
    index = 2
    disabled = False

    ngen = RefineMethodOption('Maximum # of generations', NGEN, [1, 1000], int)
    nswarms = RefineMethodOption('Start # of swarms', NSWARMS, [1, 50], int)
    nexcess = RefineMethodOption('Max # of unconverged swarms', NEXCESS, [1, 50], int)
    nparticles = RefineMethodOption('Swarm size', NPARTICLES, [1, 50], int)
    conv_factr = RefineMethodOption('Convergence tolerance', CONV_FACTR, [0., 10.], float)

    def _individual_creator(self, context, bounds):
        creator.create(
            "Particle", pyxrd_array,
            fitness=FitnessMin, #@UndefinedVariable
            speed=list,
            best=None,
            bestfit=FitnessMin, #@UndefinedVariable
            context=context,
            min_bounds=bounds[:, 0],
            max_bounds=bounds[:, 1],
        )
        creator.create("Swarm", list, best=None, bestfit=FitnessMin) #@UndefinedVariable

    def _create_stats(self):
        stats = tools.Statistics(lambda ind: ind.fitness.values)
        stats.register("avg", np.mean, axis=0)
        stats.register("std", np.std, axis=0)
        stats.register("min", np.min, axis=0)
        stats.register("max", np.max, axis=0)
        return stats

    def run(self, refiner, ngen=NGEN, nswarms=NSWARMS, nexcess=NEXCESS,
            conv_factr=CONV_FACTR, nparticles=NPARTICLES, **kwargs):

        logger.info("Setting up the DEAP MPSO refinement algorithm")

        # Process some general stuff:
        ndim = len(refiner.refinables)
        bounds = np.array(refiner.ranges)
        norms = np.abs(bounds[:, 1] - bounds[:, 0])
        self._individual_creator(refiner, bounds)

        # Strategy setup
        strategy = MultiPSOStrategy()

        # Hall of fame & stats:
        logger.info("Creating hall-off-fame and statistics")
        halloffame = PyXRDParetoFront(similar=lambda a1, a2: np.all(a1 == a2))
        stats = self._create_stats()

        # Our toolbox:
        toolbox = base.Toolbox()
        toolbox.register(
            "particle", strategy.generate_particle, creator.Particle, #@UndefinedVariable
            dim=ndim,
            pmin=bounds[:, 0], pmax=bounds[:, 1],
            smin=-norms / 2.0, smax=norms / 2.0
        )
        toolbox.register("update_particle", strategy.update_particle, chi=0.729843788, c=norms / np.amax(norms))
        toolbox.register("generate_particles", strategy.generate_particles, toolbox.particle, n=NPARTICLES)

        toolbox.register("swarm", strategy.create_swarm, creator.Swarm) #@UndefinedVariable
        toolbox.register("update_swarm", strategy.update_swarm)
        toolbox.register("give_reinit_swarms", strategy.give_reinit_swarms)
        toolbox.register("give_converged_swarms", strategy.give_converged_swarms)

        # Create algorithm
        algorithm = MPSOAlgorithm(
            toolbox, bounds, norms,
            ngen, nswarms, nexcess, nparticles, conv_factr,
            stats=stats, halloffame=halloffame,
            refiner=refiner, stop=self._stop, **kwargs
        )

        # Get this show on the road:
        logger.info("Running the MPSO algorithm...")
        best, population, converged_bests = algorithm.run() # returns (best, population) tuple

        return best, population, converged_bests

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import os, locale

import logging
logger = logging.getLogger(__name__)

from mvc.adapters.gtk_support.treemodels.utils import create_treestore_from_directory
from mvc.adapters.gtk_support.dialogs.dialog_factory import DialogFactory

from pyxrd.generic.views.cell_renderer_tools import get_default_renderer
from pyxrd.file_parsers.wld_parsers import wld_parsers
from pyxrd.file_parsers.goniometer_parsers import goniometer_parsers

from pyxrd.data import settings
from pyxrd.generic.controllers import BaseController
from pyxrd.generic.controllers.dialog_controller import DialogController
from pyxrd.generic.controllers.objectliststore_controllers import TreeViewMixin

from pyxrd.generic.views.treeview_tools import setup_treeview, new_text_column

from pyxrd.goniometer.views import WavelengthDistributionView

class InlineGoniometerController(BaseController):
    """
        Goniometer controller. Is not expected to be used with a dialog view,
        but rather in another view. 
    """

    auto_adapt_excluded = [ 'wavelength_distribution', ]


    # ------------------------------------------------------------
    #      Initialisation and other internals
    # ------------------------------------------------------------
    def register_view(self, view):
        self.generate_import_combo()
        
        self.wld_view = WavelengthDistributionView()
        self.wld_ctrl = WavelengthDistributionController(model=self.model, view=self.wld_view, parent=self)
        
        self.update_soller_sensitivity()
        self.update_divergence_label()
        self.update_absorption_sensitivity()     

    def update_soller_sensitivity(self):
        self.view["gonio_soller1_spb"].set_sensitive(self.model.has_soller1)
        self.view["gonio_soller2_spb"].set_sensitive(self.model.has_soller2)

    def update_divergence_label(self):
        if self.model.divergence_mode == "AUTOMATIC":
            self.view["unit_divergence_lbl"].set_markup("<i>[mm]</i>")
        elif self.model.divergence_mode == "FIXED":
            self.view["unit_divergence_lbl"].set_markup("<i>[°]</i>")

    def update_absorption_sensitivity(self):
        self.view["sample_surf_density_spb"].set_sensitive(self.model.has_absorption_correction)
        self.view["absorption_spb"].set_sensitive(self.model.has_absorption_correction)

    def generate_import_combo(self):
        # TODO seperate this more the gtk level...
        self.view.import_combo_box.clear()
        
        path = settings.DATA_REG.get_directory_path("DEFAULT_GONIOS")
        cmb_model = create_treestore_from_directory(path)
        self.view.import_combo_box.set_model(cmb_model)
        cell = get_default_renderer('text')
        self.view.import_combo_box.pack_start(cell, True)
        self.view.import_combo_box.add_attribute(cell, 'text', 0)
        self.view.import_combo_box.add_attribute(cell, 'sensitive', 2)

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_btn_export_gonio_clicked(self, widget, *args):
        def on_accept(dialog):
            dialog.parser.write(self.model, dialog.filename)
            self.generate_import_combo()
        DialogFactory.get_save_dialog(
            title="Select the goniometer setup file to save to",
            filters=goniometer_parsers.get_export_file_filters(),
            current_folder=settings.DATA_REG.get_directory_path("DEFAULT_GONIOS"),
            parent=self.view.parent.get_top_widget()
        ).run(on_accept)

    def on_cmb_import_gonio_changed(self, combobox, *args):
        model = combobox.get_model()
        itr = combobox.get_active_iter()
        if itr:
            # first column is the name, second column the path and third column
            # a flag indicating if this can be selected
            path = model.get_value(itr, 1)
            if path:
                def on_accept(dialog):
                    self.model.reset_from_file(path)
                DialogFactory.get_confirmation_dialog(
                    "Are you sure?\nYou will loose the current settings!",
                    parent=self.view.get_toplevel()
                ).run(on_accept)
        combobox.set_active(-1) # deselect

    def on_btn_edit_wld_clicked(self, widget, *args):
        self.wld_view.show_all()

    # ------------------------------------------------------------
    #      Notifications of observable properties
    # ------------------------------------------------------------
    @BaseController.observe("has_soller1", assign=True, after=True)
    def notif_has_soller1(self, model, prop_name, info):
        self.update_soller_sensitivity()

    @BaseController.observe("has_soller2", assign=True, after=True)
    def notif_has_soller2(self, model, prop_name, info):
        self.update_soller_sensitivity()
    
    @BaseController.observe("divergence_mode", assign=True, after=True)
    def notif_divergence(self, model, prop_name, info):
        self.update_divergence_label()
        
    @BaseController.observe("has_absorption_correction", assign=True, after=True)
    def notif_absorption(self, model, prop_name, info):
        self.update_absorption_sensitivity()

    pass # end of class

class WavelengthDistributionController(DialogController, TreeViewMixin):
    """
        Wavelength distribution controller.
    """
    
    auto_adapt_included = [ "wavelength_distribution", ]
 
    widget_handlers = {
        'custom':  'custom_handler',
    }

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    @staticmethod
    def custom_handler(self, intel, widget):
        print("CUSTOM HANDLER CALLED FOR %s" % intel.name)
 
    # ------------------------------------------------------------
    #      Initialisation and other internals
    # ------------------------------------------------------------
    def setup_wavelength_distribution_tree_view(self, store, widget):
        """
            Creates the wavelength distribution TreeView layout and behavior
        """      
        setup_treeview(widget, store,
            on_cursor_changed=self.on_wld_tv_cursor_changed,
            sel_mode='MULTIPLE')
        widget.set_model(store)
        
        def data_func(col, cell, model, iter, colnr):
            cell.set_property("text", "%g" % model.get(iter, colnr)[0])
            
        # X Column:
        widget.append_column(new_text_column(
            'Wavelength (nm)', text_col=store.c_x, editable=True,
            data_func = (data_func, (store.c_x,)),
            edited_callback=(self.on_xy_data_cell_edited, (self.model.wavelength_distribution, 0))))
        # Y Column:
        widget.append_column(new_text_column(
            'Fraction', text_col=store.c_y, editable=True,
            data_func = (data_func, (store.c_y,)),
            edited_callback=(self.on_xy_data_cell_edited, (self.model.wavelength_distribution, 1))))   
         
    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_xy_data_cell_edited(self, cell, path, new_text, model, col):
        try:
            value = float(locale.atof(new_text))
        except ValueError:
            logger.exception("ValueError: Invalid literal for float(): '%s'" % new_text)
        else:
            model.set_value(int(path), col, value)
        return True
        
    def on_wld_tv_cursor_changed(self, tv):
        path, col = tv.get_cursor()  # @UnusedVariable
        self.view["btn_del_wavelength_distribution"].set_sensitive(path is not None)
        return True
    
    def on_add_wavelength_distribution_clicked(self, widget):
        self.model.wavelength_distribution.append(0, 0)
        return True
    
    def on_del_wavelength_distribution_clicked(self, widget):
        paths = self.get_selected_paths(self.view["wld_wavelength_distribution"])
        if paths is not None:
            self.model.wavelength_distribution.remove_from_indeces(*paths)
        return True
    
    def on_import_wavelength_distribution_clicked(self, widget, data=None):
        def on_confirm(dialog):
            def on_accept(dialog):
                filename = dialog.filename
                parser = dialog.parser
                message = "An unexpected error has occurred when trying to parse %s:\n\n<i>" % os.path.basename(filename)
                message += "{}</i>\n\n"
                message += "This is most likely caused by an invalid or unsupported file format."

                with DialogFactory.error_dialog_handler(message, parent=self.view.get_toplevel(), reraise=False):
                    self.model.wavelength_distribution.load_data(parser, filename, clear=True)
            DialogFactory.get_load_dialog(
                title="Import wavelength distribution", parent=self.view.get_top_widget(),
                filters=wld_parsers.get_import_file_filters(),
                current_folder=settings.DATA_REG.get_directory_path("DEFAULT_WL_DISTR")
            ).run(on_accept)
        DialogFactory.get_confirmation_dialog(
            "Importing a wavelength distribution will erase all current data.\nAre you sure you want to continue?",
            parent=self.view.get_top_widget()
        ).run(on_confirm)
    
    def on_export_wavelength_distribution_clicked(self, widget, data=None):
        def on_accept(dialog):
            filename = dialog.filename
            parser = dialog.parser
            message = "An unexpected error has occurred when trying to save '%s'.\n" % os.path.basename(filename)
            message += "Contact the developer about this!"
            with DialogFactory.error_dialog_handler(message, parent=self.view.get_toplevel(), reraise=False):
                header = "%s, %s" % ("Wavelength", "Factor")
                self.model.wavelength_distribution.save_data(parser, filename, header=header)
        DialogFactory.get_save_dialog(
            "Select file for wavelength distribution export",
            parent=self.view.get_top_widget(), 
            filters=wld_parsers.get_export_file_filters(),
            current_folder=settings.DATA_REG.get_directory_path("DEFAULT_WL_DISTR")
        ).run(on_accept)
    
    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk
from .basic import GtkAdapter

class TextViewAdapter(GtkAdapter):
    """
        An adapter for a TextView widget.
    """
    widget_types = ["text_view", ]
    _check_widget_type = Gtk.TextView

    _prop_cast = False

    def _read_widget(self):
        """Returns the value currently stored into the widget."""
        return str(self._buffer.get_text(*self._buffer.get_bounds()))

    def _write_widget(self, val):
        """Writes value into the widget. If specified, user setter
        is invoked."""
        with self._ignore_notifications():
            return self._buffer.set_text(val)


    _signal = "changed"

    def __init__(self, controller, prop, widget,
                 value_error=None, spurious=False, update=True):

        if prop.data_type == object: # assume TextBuffer type
            # TODO
            self._buffer = self._read_property()
        else: # assume string type
            self._buffer = Gtk.TextBuffer()

        super(TextViewAdapter, self).__init__(controller, prop, widget,
                 value_error=value_error, spurious=spurious, update=update)

        self._widget.set_buffer(self._buffer)

    def _connect_widget(self):
        """Called when the adapter is ready to connect to the widget"""

        # Connect the widget
        if self._signal:
            self._signal_id = self._buffer.connect(
               self._signal, self._on_wid_changed, self._signal_args)

        # Updates the widget:
        if self._update: self.update_widget()
        return

    def _disconnect_widget(self, widget=None):
        """Disconnects the widget"""
        if self._signal is not None and self._signal_id is not None:
            self._buffer.disconnect(self._signal_id)

    def _set_property_value(self, val):
        """Private method that sets the property value stored in the model,
        without transformations."""
        return setattr(self._model, self._prop.label, val)

    pass # end of class




# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pkg_resources import resource_filename # @UnresolvedImport

from pyxrd.generic.views import DialogView

class ProjectView(DialogView):
    title = "Edit Project"
    subview_builder = resource_filename(__name__, "glade/project.glade")
    subview_toplevel = "nbk_edit_project"
    resizable = False

    widget_format = "project_%s"

    widget_groups = {
        'full_mode_only': [
            "algn_calc_color",
            "lbl_calccolor",
            "algn_calc_lw",
            "calc_lw_lbl"
        ]
    }

    @property
    def specimens_treeview_container(self):
        return self["vbox_specimens"]

    @property
    def specimens_treeview(self):
        return self["project_specimens"]

    def __init__(self, *args, **kwargs):
        DialogView.__init__(self, *args, **kwargs)
        self["popup_menu_item_add_specimen"].set_related_action(self.parent["add_specimen"])
        self["popup_menu_item_edit_specimen"].set_related_action(self.parent["edit_specimen"])
        self["popup_menu_item_import_specimens"].set_related_action(self.parent["import_specimens"])
        self["popup_menu_item_replace_data"].set_related_action(self.parent["replace_specimen_data"])
        self["popup_menu_item_export_data"].set_related_action(self.parent["export_specimen_data"])
        self["popup_menu_item_del_specimen"].set_related_action(self.parent["del_specimen"])

    def present(self, *args, **kwargs):
        super(ProjectView, self).present(*args, **kwargs)
        self["nbk_edit_project"].set_current_page(0)

    def show_specimens_context_menu(self, event):
        self["specimen_popup"].popup(None, None, None, None, event.button, event.time)

    def hide_specimens_context_menu(self):
        self["specimen_popup"].hide()

    def set_x_range_sensitive(self, sensitive):
        self["box_manual_xrange"].set_sensitive(sensitive)

    def set_y_range_sensitive(self, sensitive):
        self["box_manual_yrange"].set_sensitive(sensitive)

    def set_selection_state(self, value):
        if value is None:
            self.hide_specimens_context_menu()
        pass

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

class DataMixin(object):
    """
    Mixing for the ~:class:`mvc.models.properties.LabeledProperty` descriptor
    that allows this property to be set on the `data_object` object of the
    instance this property belongs to, instead of a private attribute.
    
    When this Mixin is used, the user can pass an additional keyword 
    argument to the descriptor:
        - data_object_label: the private attribute label for the data object,
          defaults to '_data_object' 
    """

    data_object_label = "_data_object"

    def _get_private_label(self):
        """ Private attribute label (holds the actual value on the model) """
        return "%s.%s" % (
            self.data_object_label,
            self.label
        )

    pass


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pyxrd.generic.io import get_case_insensitive_glob

from pyxrd.file_parsers.json_parser import JSONParser

from .namespace import project_parsers

@project_parsers.register_parser()
class JSONProjectParser(JSONParser):

    description = "Project file *.PYXRD"
    extensions = get_case_insensitive_glob("*.PYXRD")
    mimetypes = ["application/octet-stream", "application/zip"]


    @classmethod
    def _parse_data(cls, filename, fp, data_objects=None, close=True):
        project = super(JSONProjectParser, cls)._parse_data(filename, fp, data_objects=data_objects, close=close)
        project.filename = filename
        return project

    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.



import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

from math import sqrt

import numpy as np
import scipy

from deap import cma, base, creator, tools #@UnresolvedImport

from pyxrd.refinement.refine_method import RefineMethod
from pyxrd.refinement.refine_method_option import RefineMethodOption
from pyxrd.refinement.refine_async_helper import RefineAsyncHelper

from .deap_utils import pyxrd_array, PyXRDParetoFront, FitnessMin, result_func

# Default settings:
NGEN = 100
STAGN_NGEN = 10
STAGN_TOL = 0.001

class Strategy(cma.Strategy):
    """
        This evolutionary strategy supports the hybrid PSO-CMA runs using the
        rotate_and_bias function (should be called after an update).
    """

    def __init__(self, centroid, sigma, ranges, **kwargs):
        self.ranges = ranges
        super(Strategy, self).__init__(centroid, sigma, **kwargs)

    def update(self, population):
        """Update the current covariance matrix strategy from the
        *population*.
        
        :param population: A list of individuals from which to update the
                           parameters.
        """
        population.sort(key=lambda ind: ind.fitness, reverse=True)
        selected_pop = self._translate_external(
            np.array([ind.to_ndarray() for ind in population[0:self.mu]]))

        old_centroid = self._translate_external(self.centroid)
        centroid = np.dot(self.weights, selected_pop)

        c_diff = centroid - old_centroid

        # Cumulation : update evolution path
        self.ps = (1 - self.cs) * self.ps \
             + sqrt(self.cs * (2 - self.cs) * self.mueff) / self.sigma \
             * np.dot(self.B, (1. / self.diagD) \
                          * np.dot(self.B.T, c_diff))

        hsig = float((np.linalg.norm(self.ps) /
                sqrt(1. - (1. - self.cs) ** (2. * (self.update_count + 1.))) / self.chiN
                < (1.4 + 2. / (self.dim + 1.))))

        self.update_count += 1

        self.pc = (1 - self.cc) * self.pc + hsig \
                  * sqrt(self.cc * (2 - self.cc) * self.mueff) / self.sigma \
                  * c_diff

        # Update covariance matrix
        artmp = selected_pop - old_centroid
        new_C = (1 - self.ccov1 - self.ccovmu + (1 - hsig) \
                   * self.ccov1 * self.cc * (2 - self.cc)) * self.C \
                + self.ccov1 * np.outer(self.pc, self.pc) \
                + self.ccovmu * np.dot((self.weights * artmp.T), artmp) \
                / self.sigma ** 2

        self.sigma *= np.exp((np.linalg.norm(self.ps) / self.chiN - 1.) \
                                * self.cs / self.damps)

        try:
            self.diagD, self.B = np.linalg.eigh(new_C)
        except np.linalg.LinAlgError:
            logger.warning(
                "LinAlgError occurred when calculating eigenvalues" \
                " and vectors for matrix C!\n%r" % new_C
            )
        else:
            self.C = new_C
            indx = np.argsort(self.diagD)

            self.cond = self.diagD[indx[-1]] / self.diagD[indx[0]]

            self.diagD = self.diagD ** 0.5

            self.B = self.B[:, indx]
            self.BD = self.B * self.diagD

        self.centroid = self._translate_internal(centroid)

    def rotate_and_bias(self, global_best, tc=0.1, b=0.5, cp=0.5):
        """
            Rotates the covariance matrix and biases the centroid of this
            CMA population towards a global mean. Can be used to implement a
            PSO-CMA hybrid algorithm. 
        """

        global_best = self._translate_external(global_best)
        centroid = self._translate_external(self.centroid)

        # Rotate towards global:
        pg = np.array(global_best) - np.array(centroid)
        Brot = self.__rotation_matrix(self.B[:, 0], pg) * self.B
        Crot = Brot * (self.diagD ** 2) * Brot.T
        self.C = cp * self.C + (1.0 - cp) * Crot

        # Bias our mean towards global best mean:
        npg = np.linalg.norm(pg)
        nsigma = np.amax(self.sigma)
        if nsigma < npg:
            if nsigma / npg <= tc * npg:
                bias = b * pg
            else:
                bias = nsigma / npg * pg
        else:
            bias = 0

        centroid = centroid + bias

        self.centroid = self._translate_internal(centroid)

        pass

    def _translate_internal(self, solutions):
        # rule is: anything given as an argument in a public function or
        # available as a public property should be within the external boundaries

        return self.ranges[:, 0] + (self.ranges[:, 1] - self.ranges[:, 0]) * (1.0 - np.cos(solutions * np.pi)) / 2.0

    def _translate_external(self, solutions):
        return np.arccos(1 - 2 * (solutions - self.ranges[:, 0]) / (self.ranges[:, 1] - self.ranges[:, 0])) / np.pi

    def generate(self, ind_init):
        """Generate a population from the current strategy using the 
        centroid individual as parent.
        
        :param ind_init: A function object that is able to initialize an
                         individual from a list.
        :returns: an iterator yielding the generated individuals.
        """

        centroid = self._translate_external(self.centroid)

        arz = np.random.standard_normal((self.lambda_, self.dim)) #@UndefinedVariable
        arz = np.array(centroid) + self.sigma * np.dot(arz, self.BD.T) #@UndefinedVariable

        arz = self._translate_internal(arz)

        for arr in arz:
            yield ind_init(arr)

    def __rotation_matrix(self, vector, target):
        """ Rotation matrix from one vector to another target vector.
     
        The solution is not unique as any additional rotation perpendicular to
        the target vector will also yield a solution)
         
        However, the output is deterministic.
        """

        R1 = self.__rotation_to_pole(target)
        R2 = self.__rotation_to_pole(vector)

        return np.dot(R1.T, R2)

    def __rotation_to_pole(self, target):
        """ Rotate to 1,0,0... """
        n = len(target)
        working = target
        rm = np.eye(n)
        for i in range(1, n):
            angle = np.arctan2(working[0], working[i])
            rm = np.dot(self.__rotation_matrix_inds(angle, n, 0, i), rm)
            working = np.dot(rm, target)

        return rm

    def __rotation_matrix_inds(self, angle, n, ax1, ax2):
        """ 'n'-dimensional rotation matrix 'angle' radians in coordinate plane with
            indices 'ax1' and 'ax2' """


        s = np.sin(angle)
        c = np.cos(angle)

        i = np.eye(n)

        i[ax1, ax1] = s
        i[ax1, ax2] = c
        i[ax2, ax1] = c
        i[ax2, ax2] = -s

        return i

    pass #end of class

class Algorithm(RefineAsyncHelper):
    """
        This algorithm implements the ask-tell model proposed in 
        [Colette2010]_, where ask is called `generate` and tell is called `update`.
        
        Modified (Mathijs Dumon) so it checks for stagnation.
    """

    @property
    def ngen(self):
        return self._ngen
    @ngen.setter
    def ngen(self, value):
        self._ngen = value
        logger.info("Setting ngen to %d" % value)
    _ngen = 100

    gen = -1
    halloffame = None

    refiner = None

    toolbox = None
    stats = None
    stagn_ngen = None
    stagn_tol = None
    verbose = False

    #--------------------------------------------------------------------------
    #    Initialization
    #--------------------------------------------------------------------------
    def __init__(self, toolbox, halloffame, stats, ngen=NGEN,
                         verbose=__debug__, stagn_ngen=STAGN_NGEN,
                         stagn_tol=STAGN_TOL, refiner=None, stop=None):
        """
        :param toolbox: A :class:`~deap.base.Toolbox` that contains the evolution
                        operators.
        :param ngen: The number of generations.
        :param halloffame: A :class:`~deap.tools.ParetoFront` object that will
                           contain the best individuals.
        :param stats: A :class:`~deap.tools.Statistics` object that is updated
                      inplace.
        :param verbose: Whether or not to log the statistics.
        :param stagn_gens: The minimum number of generations to wait before
                            checking for stagnation
        :param stagn_tol: The stagnation tolerance. Higher values means a 
                            harsher tolerance, values should fall between 0 and 1
                    
        :param refiner: PyXRD refiner object
    
        :returns: The best individual and the final population.
        
        The toolbox should contain a reference to the generate and the update method 
        of the chosen strategy.
        
        Call the run() method when the algorithm should be run.
    
        .. [Colette2010] Collette, Y., N. Hansen, G. Pujol, D. Salazar Aponte and
           R. Le Riche (2010). On Object-Oriented Programming of Optimizers -
           Examples in Scilab. In P. Breitkopf and R. F. Coelho, eds.:
           Multidisciplinary Design Optimization in Computational Mechanics,
           Wiley, pp. 527-565;
    
        """
        self.stats = stats
        self.toolbox = toolbox
        self.ngen = ngen
        self.halloffame = halloffame
        self.verbose = verbose
        self.stagn_ngen = stagn_ngen
        self.stagn_tol = stagn_tol
        self.refiner = refiner

        self.gen = 0

        self._stop = stop

    #--------------------------------------------------------------------------
    #    Run method:
    #--------------------------------------------------------------------------
    def run(self):
        """Will run this algorithm"""
        if self.verbose:
            column_names = ["gen", "evals", "best"]
            if self.stats is not None:
                column_names += list(self.stats.functions.keys())
            self.logbook = tools.Logbook()
            self.logbook.header = column_names

        for _ in range(self.ngen):
            # Check if the user has cancelled:
            if self._user_cancelled():
                self.refiner.status.message = "Stopping..."
                logger.info("User cancelled execution, stopping ...")
                break

            #ASK: Generate a new population:
            population = self._ask()
            #TELL: Update the strategy with the evaluated individuals
            self._tell(population)
            #RECORD: For stagnation checking & logging:
            self._record(population)
            #CHECK: whether we are stagnating:
            if self._is_stagnating():
                logging.info("CMA: stagnation detected!")
                break

        return self.refiner.history.best_solution, population

    #--------------------------------------------------------------------------
    #    Stagnation calls:
    #--------------------------------------------------------------------------
    def _is_flat(self, yvals, xvals, slope_tolerance=0.001):
        slope, intercept, r_value, p_value, std_err = scipy.stats.linregress(xvals, yvals) #@UndefinedVariable @UnusedVariable
        val = bool(abs(slope) <= slope_tolerance)
        return val

    def _is_stagnating(self):
        self.refiner.status.message = "Checking for stagnation"
        if self.gen >= self.stagn_ngen: # 10
            std, best = self.logbook.select("std", "best")
            std = np.array(std)[:, 0]
            yvals1 = std[-(self.stagn_ngen - 1):]
            xvals1 = list(range(len(yvals1)))
            yvals2 = best[-(self.stagn_ngen - 1):]
            xvals2 = list(range(len(yvals2)))
            return self._is_flat(yvals1, xvals1, self.stagn_tol) and \
                self._is_flat(yvals2, xvals2, self.stagn_tol)
        else:
            return False

    #--------------------------------------------------------------------------
    #    Ask, tell & record:
    #--------------------------------------------------------------------------
    def _ask(self):
        self.gen += 1
        self.refiner.status.message = "Creating generation #%d" % self.gen

        def result_f(*args):
            self.refiner.update(*args)
            result_func(*args)

        population = self.do_async_evaluation(
            self.toolbox.generate, result_func=result_f
        )
        
        if self.halloffame is not None:
            self.halloffame.update(population)

        return population

    def _tell(self, population):
        self.refiner.status.message = "Updating strategy"
        self.toolbox.update(population)

    def _record(self, population):
        self.refiner.status.message = "Processing ..."
        
        # Get the best solution so far:
        best = self.halloffame.get_best()
        best_f = best.fitness.values[0]
        pop_size = len(population)

        # Calculate stats & print something if needed:
        record = self.stats.compile(population)
        if self.verbose:
            self.logbook.record(gen=self.gen, evals=pop_size, best=best_f, **record)
            print(self.logbook.stream)

        self.refiner.status.message = "Refiner update ..."
        # Update the refiner:
        self.refiner.update(best, iteration=self.gen, residual=best_f)

    pass #end of class

class RefineCMAESRun(RefineMethod):
    """
        The DEAP CMA-ES algorithm implementation with added stagnation thresholds
    """
    name = "CMA-ES refinement"
    description = "This algorithm uses the CMA-ES refinement strategy as implemented by DEAP"
    index = 1
    disabled = False

    ngen = RefineMethodOption('Maximum # of generations', NGEN, [1, 10000], int)
    stagn_ngen = RefineMethodOption('Minimum # of generations', STAGN_NGEN, [1, 10000], int)
    stagn_tol = RefineMethodOption('Fitness slope tolerance', STAGN_TOL, [0., 100.], float)

    def _individual_creator(self, refiner, bounds):
        creator.create(
            "Individual", pyxrd_array,
            fitness=FitnessMin, # @UndefinedVariable
            refiner=refiner,
            min_bounds=bounds[:, 0].copy(),
            max_bounds=bounds[:, 1].copy(),
        )

        def create_individual(lst):
            arr = np.array(lst).clip(bounds[:, 0], bounds[:, 1]) #@UndefinedVariable
            return creator.Individual(arr) # @UndefinedVariable

        return create_individual

    def _create_stats(self):
        stats = tools.Statistics(lambda ind: ind.fitness.values)
        stats.register("avg", np.mean, axis=0) #@UndefinedVariable
        stats.register("std", np.std, axis=0) #@UndefinedVariable
        stats.register("min", np.min, axis=0) #@UndefinedVariable
        stats.register("max", np.max, axis=0) #@UndefinedVariable
        return stats

    _has_been_setup = False
    def _setup(self, refiner, ngen=NGEN, stagn_ngen=STAGN_NGEN, stagn_tol=STAGN_TOL, **kwargs):
        if not self._has_been_setup:
            logger.info("Setting up the DEAP CMA-ES refinement algorithm (ngen=%d)" % ngen)
            refiner.status.message = "Setting up algorithm..."

            # Process some general stuff:
            bounds = np.array(refiner.ranges) #@UndefinedVariable
            create_individual = self._individual_creator(refiner, bounds)

            # Setup strategy:
            centroid = create_individual(refiner.history.initial_solution)
            strat_kwargs = {}
            if "lambda_" in kwargs:
                strat_kwargs["lambda_"] = kwargs.pop("lambda_")
            strategy = Strategy(
                centroid=centroid, sigma=1.0 / 10.0, ranges=bounds,
                stop=self._stop, **strat_kwargs
            )

            # Toolbox setup:
            toolbox = base.Toolbox()
            toolbox.register("generate", strategy.generate, create_individual)
            toolbox.register("update", strategy.update)

            # Hall of fame & stats:
            logger.info("Creating hall-off-fame and statistics")
            halloffame = PyXRDParetoFront(similar=lambda a1, a2: np.all(a1 == a2)) #@UndefinedVariable
            stats = self._create_stats()

            # Create algorithm
            self.algorithm = Algorithm(
                toolbox, halloffame, stats, ngen=ngen,
                stagn_ngen=stagn_ngen, stagn_tol=stagn_tol, refiner=refiner, stop=self._stop)

            self._has_been_setup = True
        return self.algorithm

    def run(self, refiner, **kwargs):
        logger.info("CMA-ES run invoked with %s" % kwargs)
        self._has_been_setup = False #clear this for a new refinement
        algorithm = self._setup(refiner, **kwargs)
        # Get this show on the road:
        logger.info("Running the CMA-ES algorithm...")
        algorithm.run()

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from contextlib import contextmanager, ExitStack

class EventContextManager(object):
    """
        Event context manager class, to be used as follows:
        
         ecm = EventContextManager(model.event1, model.event2, ...)
        
         with ecm.ignore():
             pass #do something here that will cause events to be ignored
        
         with ecm.hold():
             pass #do something here that will cause events to be held back
        
    """
    
    events = []
    
    def __init__(self, *events):
        self.events = events
        
    @contextmanager
    def ignore(self):
        if len(self.events):
            with ExitStack() as stack:
                for event in self.events:
                    stack.enter_context(event.ignore())
                yield
        else:
            yield

    @contextmanager
    def hold(self):
        if len(self.events):
            with ExitStack() as stack:
                for event in self.events:
                    stack.enter_context(event.hold())
                yield
        else:
            yield

    pass #end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

import numpy as np

from .refine_history import RefineHistory
from .refine_status import RefineStatus

class RefineSetupError(ValueError):
    """ Raised if an error exists in the refinement setup """
    pass

class Refiner(object):
    """
        A model for the refinement procedure.
    """

    method = None
    
    history = None
    status = None

    refinables = None

    def __init__(self, method, data_callback, refinables, event_cmgr, metadata={}):
        super(Refiner, self).__init__()
        
        assert method is not None, "Cannot refine without a refinement method!"
        residual_callback = method.residual_callback
        assert callable(residual_callback), "Cannot refine without a residual callback!"
        assert callable(data_callback), "Cannot refine without a data callback!"
        assert refinables is not None, "Cannot refine without refinables!"
        assert event_cmgr is not None, "Cannot refine without an event context manager!"
        
        # Set these:
        self.method = method
        self.residual_callback = residual_callback
        self.data_callback = data_callback
        self.event_cmgr = event_cmgr
        self.metadata = metadata

        # Create the refinement history object:
        logger.info("Setting up the refinement history.") 
        self.history = RefineHistory()

        # Create the refinement status object:
        logger.info("Setting up the refinement status object.")        
        self.status = RefineStatus(self.history)

        # Setup the refinable property list:
        logger.info("Refinement with the following refinables:")
        self.refinables = []
        self.ranges = ()
        self.labels = ()
        initial_values = []
        for node in refinables.iter_children():
            refinable = node.object
            if refinable.refine and refinable.refinable:
                logger.info(" - %s from %r" % (refinable.text_title, refinable.obj))
                self.refinables.append(refinable)
                initial_values.append(refinable.value)
                if not (refinable.value_min < refinable.value_max):
                    logger.info("Error in refinement setup!")
                    self.status.error = True
                    self.status.message = "Invalid parameter range for '%s'!" % (refinable.text_descriptor,)
                    raise RefineSetupError("Invalid parameter range for '%s'!" % (refinable.text_descriptor,))
                self.ranges += ((refinable.value_min, refinable.value_max),)
                self.labels += ((refinable.text_title, refinable.title),)

        # Make sure we can refine something:
        if len(self.refinables) == 0:
            logger.error("No refinables selected!")
            self.status.error = True
            self.status.message = "No parameters selected!"
            raise RefineSetupError("No parameters selected!")      

        # Register the initial solution:
        initial_solution = np.array(initial_values, dtype=float)
        self.history.set_initial_solution(
            initial_solution,
            self.get_history_residual(self.get_residual(initial_solution))
        )

    def apply_solution(self, solution):
        """
            Applies the given solution
        """
        solution = np.asanyarray(solution)
        with self.event_cmgr.hold():
            for i, ref_prop in enumerate(self.refinables):
                if not (solution.shape == ()):
                    ref_prop.value = float(solution[i])
                else:
                    ref_prop.value = float(solution[()])

    def get_data_object(self, solution):
        """
            Gets the mixture data object after setting the given solution
        """
        with self.event_cmgr.ignore():
            self.apply_solution(solution)
            return self.data_callback()
        
    def get_residual(self, solution):
        """
            Gets the residual for the given solution after setting it
        """
        return self.residual_callback(self.get_data_object(solution))

    def get_history_residual(self, residual):
        try:
            return residual[0]
        except IndexError:
            return residual 

    def update(self, solution, residual=None, iteration=0):
        """
            Update's the refinement contect with the given solution:
                - applies the solution & gets the residual if not given
                - stores it in the history
        """
        residual = self.get_history_residual(
            residual if residual is not None else self.get_residual(solution)
        )
        self.history.register_solution(iteration, solution, residual)

    def apply_best_solution(self):
        self.apply_solution(self.history.best_solution)

    def apply_last_solution(self):
        self.apply_solution(self.history.last_solution)

    def apply_initial_solution(self):
        self.apply_solution(self.history.initial_solution)
    
    def get_plot_samples(self):
        return self.history.samples[:,self.history.PLOT_SAMPLE_SELECTOR]
    
    def get_plot_labels(self):
        return [plot_label for plot_label, _ in self.labels] + ["Rp",]
    
    def refine(self, stop):       
        # Suppress updates:
        with self.event_cmgr.hold():
            # Make sure the stop signal is not set from a previous run:
            stop.clear()

            # Log some information:
            logger.info("-"*80)
            logger.info("Starting refinement with this setup:")
            msg_frm = "%22s: %s"
            logger.info(msg_frm % ("refinement method", self.method))
            logger.info(msg_frm % ("number of parameters", len(self.refinables)))

            # Run the refinement:
            with self.status:
                with self.history:
                    self.method(self, stop=stop)

            # Log some more information:
            logger.info('Total refinement took %0.3f ms' % self.status.get_total_time())
            logger.info('Best solution found was:')                
            for i, ref_prop in enumerate(self.refinables):
                logger.info("%25s: %f" % (
                    ref_prop.text_descriptor, 
                    self.history.best_solution[i]
                ))
            logger.info("-"*80)
                
            # Return us to whatever called this
            return self
        
    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .specimen_controllers import SpecimenController, StatisticsController

from .marker_controllers import (
    MarkersController,
    EditMarkerController,
    MatchMineralController,
    ThresholdController
)

from pyxrd.generic.controllers.line_controllers import (
    BackgroundController,
    SmoothDataController,
    AddNoiseController,
    ShiftDataController,
    StripPeakController
)

__all__ = [
    "SpecimenController",
    "StatisticsController",
    "MarkersController",
    "EditMarkerController",
    "MatchMineralController",
    "ThresholdController",
    "BackgroundController",
    "SmoothDataController",
    "AddNoiseController",
    "ShiftDataController",
    "StripPeakController",
]



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

"""
This module contains the basic implementation of the matrix formalism as
detailed in Drits and Tchoubar (1990) and Plançon (2001).

It was chosen to implement this using 'loose' function calls. The disadvantage
of this approach is that the functions are no longer bound to class instances,
which makes them less intuitive to use. The advantage is we can more easily
call these functions asynchronously (e.g. using :py:class:`~multiprocessing.Pool`) 

Despite all this, most function calls in this module do expect to be passed a
:class:`~pyxrd.calculations.DataObject` sub-class, which wraps all the data in a single object. 
These :class:`~pyxrd.calculations.DataObject` s map onto the different models used. As such this 
module is also largely independent from the MVC framework used.
  
Drits, V.A., and Tchoubar, C., 1990. X-Ray Diffraction by Disordered Lamellar Structures: Theory and Applications to Microdivided Silicates and Carbons. Springer-Verlag, Berlin, Germany.
Plançon, A., 2001. Order-disorder in clay mineral structures. Clay Miner 36, 1–14.

"""


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.


import time
import logging
logger = logging.getLogger(__name__)

import numpy as np
from scipy.optimize import fmin_l_bfgs_b

from pyxrd.data import settings

from .specimen import calculate_phase_intensities, calculate_scaled_intensities, get_clipped_intensities
from .exceptions import wrap_exceptions
from .statistics import Rp, Rpw, Rpder, Rphase

__residual_method_map = {
    "Rp" : Rp,
    "Rpw" : Rpw,
    "Rpder": Rpder,
    "Rphase": Rphase
}

#===============================================================================
# Solution vector handling:
#===============================================================================
def parse_solution(x, mixture):
    # Get the original solution   
    fractions = mixture.fractions
    scales = mixture.scales
    bgshifts = mixture.bgshifts
    
    # We normalize the new fractions so the sum of all fractions
    # (including static fractions) is always 1
    scaled_fracs = x[:mixture.m] * (1.0 - mixture.sum_static_fractions) / np.sum(x[:mixture.m])
    
    # Store the variables in the right positions:
    np.put(fractions, mixture.selected_fractions, scaled_fracs)
    np.put(scales, mixture.selected_scales, x[mixture.m:mixture.m + mixture.n])
    np.put(bgshifts, mixture.selected_bgshifts, x[-mixture.o:])
    
    # Return the complete arrays (including static factors)
    return fractions, scales, bgshifts

def get_solution(mixture):
    return np.concatenate((
        np.take(mixture.fractions, mixture.selected_fractions),
        np.take(mixture.scales, mixture.selected_scales),
        np.take(mixture.bgshifts, mixture.selected_bgshifts)
    ))    

def set_solution(x, mixture):
    mixture.fractions, mixture.scales, mixture.bgshifts = parse_solution(x, mixture)   

def get_zero_solution(mixture):
    # Create complete solution array
    x0 = np.ones(shape=(mixture.m+mixture.n+mixture.o,))
    # Set all bg shifts to zero   
    x0[-mixture.o:] = 0.0
    # Set all fractions to equal size (of the remainder)
    x0[:mixture.m] = 1.0 / (1.0 - mixture.sum_static_fractions)
    
    return x0

#===============================================================================
# Solution bounds handling:
#===============================================================================
def get_bounds(mixture):
    bounds = [(0., 1.) for _ in range(mixture.m)] # allow zero fractions
    bounds += [(1e-3, None) for _ in range(mixture.n)] # don't allow zero scales
    bounds += [(0., None) for _ in range(mixture.o)]
    return bounds

#===============================================================================
# Residual calculations:
#===============================================================================
def _get_specimen_residual(specimen):
    """
        Returns the residual error for the given specimen and the (otionally)
        given calculated data. If no calculated data is passed, the calculated
        data stored in the specimen object is used (and assumed to be set).
    """
    exp, cal = get_clipped_intensities(specimen)
    return __residual_method_map[settings.RESIDUAL_METHOD](exp, cal)   

def _get_residuals(x, mixture):
    set_solution(x, mixture)
    mixture = calculate_mixture(mixture)
    return mixture.residuals

def get_residual(mixture):
    return _get_residuals(get_solution(mixture), mixture)

#===============================================================================
# Mixture calculations:
#===============================================================================

def parse_mixture(mixture, force=False):
    """
        This calculates the phase intensities.
    """
    if not mixture.parsed or force:
        mixture.parsed = False
        # Sanity checks:
        assert mixture.n > 0, "Need at least 1 specimen to optimize phase fractions, scales and background."
        assert mixture.n == len(mixture.specimens), "Invalid specimen count on mixture data object."
        assert mixture.m > 0, "Need at least 1 phase in one of the specimen to optimize phase fractions, scales and background."
        
        # We have some fixed fractions:
        mixture.selected_fractions = np.asanyarray(np.nonzero(mixture.fractions_mask)).flatten()
        mixture.selected_scales = np.asanyarray(np.nonzero(mixture.scales_mask)).flatten()
        mixture.selected_bgshifts = np.asanyarray(np.nonzero(mixture.bgshifts_mask)).flatten()
        mixture.real_m = mixture.m
        mixture.real_n = mixture.n
        mixture.real_o = mixture.n
        mixture.m = len(mixture.selected_fractions)
        mixture.n = len(mixture.selected_scales)
        mixture.o = len(mixture.selected_bgshifts)
        
        # The first term should be 1.0, unless the user entered a custom fraction
        sm = np.sum(mixture.fractions)
        mixture.sum_static_fractions = sm - np.sum(np.take(mixture.fractions, mixture.selected_fractions))
        if sm != 1.0 and mixture.sum_static_fractions < 0.0 or mixture.sum_static_fractions > 1.0:
            mixture.fractions = mixture.fractions / sm
            sm = 1.0
        mixture.sum_static_fractions = sm - np.sum(np.take(mixture.fractions, mixture.selected_fractions))
            
        
        mixture.z = 0
        for specimen in mixture.specimens:
            if specimen is not None:
                mixture.z = len(specimen.z_list)
                break
        assert mixture.z > 0, "Need at least 1 pattern in one of the specimens to optimize phase fractions, scales and background."

        for specimen in mixture.specimens:
            if specimen is not None:
                specimen.correction, specimen.phase_intensities = \
                    calculate_phase_intensities(specimen)
            else:
                logger.warning("parse_mixture reports: 'None found!'")
                
        mixture.parsed = True

@wrap_exceptions
def optimize_mixture(mixture, force=False):
    """
        Optimizes the mixture fractions, scales and bg shifts.
        Returns the mixture data object.
    """
    if not mixture.optimized or force:
        # 0. Calculate phase intensitities
        try:
            parse_mixture(mixture)
        except AssertionError:
            if settings.DEBUG:
                raise
            return mixture # ignore and return the original object back
    
        # 1. setup start point:
        x0 = get_zero_solution(mixture)
        bounds = get_bounds(mixture)
    
        # 2. Define target function:        
        def _get_average_residual(x):
            # Reset this flag so we actually recalculated stuff: 
            mixture.calculated = False
            res = _get_residuals(x, mixture)[0]
            return res
    
        # 3. Optimize:
        t1 = time.time()
        lastx, residual, info = fmin_l_bfgs_b(
            _get_average_residual,
            x0,
            approx_grad=True,
            bounds=bounds,
            iprint=-1
        )
        if np.isscalar(residual): # Make sure this is an array:
            residual = np.array([residual])
    
        t2 = time.time()
        logger.debug('%s took %0.3f ms' % ("optimize_mixture", (t2 - t1) * 1000.0))
        logger.debug(' Solution: %s' % lastx)
        logger.debug(' Average residual: %s' % residual)
        logger.debug(' Info dict: %s' % info)
    
        # 4. rescale scales and fractions so they fit into [0-1] range,
        #    and round them to have 6 digits max:
        fractions, scales, bgshifts = parse_solution(lastx, mixture)
        fractions = fractions.flatten()
    
        sum_frac = np.sum(fractions)
        if sum_frac == 0.0 and len(fractions) > 0: # prevent NaN errors
            fractions[0] = 1.0
            sum_frac = 1.0
        fractions = np.around((fractions / sum_frac), 6)
        scales *= sum_frac
        scales = scales.round(6)
    
        # 5. Set properties on data object 
        mixture.fractions = fractions
        mixture.scales = scales
        mixture.bgshifts = bgshifts
        mixture.residual = residual

        # We still need to recalculated with the last solution
        mixture.calculated = False 
        
        # But we do have optimized fractions etc.
        mixture.optimized = True

    return mixture

@wrap_exceptions
def calculate_mixture(mixture, force=False):
    """
        Calculates total intensities for the current mixture, without optimizing
        fractions, scales & background shifts.
        Returns the mixture data object.
    """
    if not mixture.calculated or force:
        try:
            parse_mixture(mixture)
        except AssertionError:
            for specimen in mixture.specimens:
                if specimen is not None:
                    specimen.total_intensity = None # clear pattern
                    specimen.scaled_phase_intensities = None # clear pattern
            return mixture
        
        fractions = np.asanyarray(mixture.fractions)
    
        # This will contain the following residuals:
        # Average, Specimen1, Specimen2, ...
        mixture.residuals = [0.0, ]
        for scale, bgshift, specimen in zip(mixture.scales, mixture.bgshifts, mixture.specimens):
            specimen_residual = 0.0
            if specimen is not None:
                bgshift = bgshift if settings.BGSHIFT else 0.0
                specimen = calculate_scaled_intensities(specimen, scale, fractions, bgshift)
                if specimen.observed_intensity.size > 0:
                    specimen_residual = _get_specimen_residual(specimen)
                else:
                    logger.warning("calculate_mixture reports: 'Zero observations found!'")
            mixture.residuals.append(specimen_residual)
        mixture.residuals[0] = np.average(mixture.residuals[1:])
        
        mixture.calculated = True
        
    return mixture

@wrap_exceptions
def calculate_and_optimize_mixture(mixture):
    """
        Calculates total intensities for the current mixture, after optimizing
        fractions, scales & background shifts.
        Returns the mixture data object.
    """
    return calculate_mixture(optimize_mixture(mixture))

@wrap_exceptions
def get_optimized_residual(mixture):
    """
        Calculates total intensities for the current mixture, after optimizing
        fractions, scales & background shifts.
        Returns the average residual instead of the mixture data object.
    """
    mixture = calculate_and_optimize_mixture(mixture)
    return mixture.residual
    


# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2016 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import unittest

from mvc.adapters.gtk_support.dialogs.dialog_factory import DialogFactory
from .file_chooser_dialog_args import get_file_chooser_kwags

__all__ = [
    'DialogFactoryTest',
]

class DialogFactoryTest(unittest.TestCase):


    def setUp(self):
        pass


    def tearDown(self):
        pass


    def test_get_file_dialog(self):
        kwargs = get_file_chooser_kwags()
        dialog = DialogFactory.get_file_dialog(**kwargs)

        self.assertEqual(dialog.get_action(), kwargs["action"], "Action attribute is not set correctly")
        self.assertEqual(dialog.get_title(), kwargs["title"], "Title attribute is not set correctly")
        self.assertEqual(dialog.get_parent(), kwargs["parent"], "Parent window is not set correctly")
        self.assertEqual(dialog.get_current_name(), kwargs["current_name"], "Current name attribute is not set correctly")
        self.assertEqual(dialog.get_current_folder(), kwargs["current_folder"], "Current folder attribute is not set correctly")
        self.assertEqual(dialog.get_extra_widget(), kwargs["extra_widget"], "Extra widget attribute is not set correctly")
        self.assertEqual(dialog.filters, kwargs["filters"], "Filters attribute is not set correctly")
        self.assertEqual(dialog.get_select_multiple(), kwargs["multiple"], "Multiple attribute is not set correctly")
        self.assertEqual(dialog.get_do_overwrite_confirmation(), kwargs["confirm_overwrite"], "Confirm overwrite attribute is not set correctly")

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import warnings
import os, sys
import logging
logger = logging.getLogger(__name__)

from mvc.support.gui_loop import start_event_loop

def _run_user_script():
    """
        Runs the user script specified in the command-line arguments.
    """
    
    available, nv, cv = check_for_updates()
    if available:
        print("An update is available: current version is %s upstream version is %s - consider upgrading!" % cv, nv)
    
    from pyxrd.data import settings

    try:
        import imp
        user_script = imp.load_source('user_script', settings.ARGS.script)
    except BaseException as err:
        err.args = "Error when trying to import %s: %s" % (settings.ARGS.script, err.args)
        raise
    user_script.run(settings.ARGS)

def _run_gui(project=None):

    # Display a splash screen showing the loading status...
    from pkg_resources import resource_filename # @UnresolvedImport
    from pyxrd.application.splash import SplashScreen
    from pyxrd import __version__

    filename = resource_filename(__name__, "application/icons/pyxrd.png")
    splash = SplashScreen(filename, __version__)

    # Run GUI:
    splash.set_message("Checking for updates ...")   
    update_available, nv, cv = check_for_updates()

    splash.set_message("Loading GUI ...")

    # Now we can load these:
    from pyxrd.data import settings
    from pyxrd.file_parsers.json_parser import JSONParser
    from pyxrd.application.models import AppModel
    from pyxrd.application.views import AppView
    from pyxrd.application.controllers import AppController
    
    # TODO move this to mvc:
    from pyxrd.generic.gtk_tools.gtkexcepthook import plugin_gtk_exception_hook

    filename = settings.ARGS.filename #@UndefinedVariable

    # Check if a filename was passed, if so try to load it
    if filename != "":
        try:
            logging.info("Opening project: %s" % filename)
            project = JSONParser.parse(filename)
        except IOError:
            logging.info("Could not load project file %s: IOError" % filename)
            # FIXME the user should be informed of this in a dialog...

    # Disable unity overlay scrollbars as they cause bugs with modal windows
    os.environ['LIBOVERLAY_SCROLLBAR'] = '0'
    os.environ['UBUNTU_MENUPROXY'] = ""

    if not settings.DEBUG:
        warnings.filterwarnings(action='ignore', category=Warning)

    # Close splash screen
    if splash: 
        splash.close()

    # Nice GUI error handler:
    gtk_exception_hook = plugin_gtk_exception_hook()

    # setup MVC:
    m = AppModel(project=project)
    v = AppView()
    AppController(m, v, gtk_exception_hook=gtk_exception_hook)

    # Free this before continuing
    del splash

    if update_available:
        from mvc.adapters.gtk_support.dialogs.dialog_factory import DialogFactory
        DialogFactory.get_information_dialog(
            "An update is available (%s) - consider upgrading!" % nv, 
            False, v.get_toplevel(), title = "Update available"
        ).run()
    else:
        print("PyXRD is up to date (current = %s)" % cv)

    # lets get this show on the road:
    start_event_loop()

def check_for_updates():
    """
        Checks for updates and returns a tuple:
            update_available, latest_version, current_version
    """
    from pyxrd.generic.outdated import check_outdated
    from pyxrd.__version import __version__
    is_outdated, latest_version = check_outdated('pyxrd', __version__)
    
    return is_outdated, latest_version, __version__

def run_main():
    """
        Parsers command line arguments and launches PyXRD accordingly.
    """

    # Make sure the current path is used for loading PyXRD modules:
    mod = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    if not mod in sys.path:
        sys.path.insert(1, mod)

    # Init settings, first import will trigger initialization
    from pyxrd.data import settings
    settings.initialize()

    # Setup basic logging
    from pyxrd.logs import setup_logging
    setup_logging(basic=True)

    if settings.DEBUG:
        from pyxrd import stacktracer
        stacktracer.trace_start(
            "trace.html",
            interval=5, auto=True) # Set auto flag to always update file!

    try:
        if settings.ARGS.script:
            # Run the specified user script:
            _run_user_script()
        else:
            # Run the GUI:
            _run_gui()
    except:
        raise # re-raise the error
    finally:
        for finalizer in settings.FINALIZERS:
            finalizer()
        if settings.DEBUG: stacktracer.trace_stop()

if __name__ == "__main__":
    run_main()



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#
# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from mvc.models.properties import StringProperty

from pyxrd.generic.io.custom_io import storables, Storable
from pyxrd.generic.models.base import DataModel

from pyxrd.refinement.refinables.mixins import RefinementGroup

@storables.register()
class InSituBehaviour(DataModel, RefinementGroup, Storable):
    """
        Interface class for coding in-situ behaviour scripts.
        Sub-classes should override or implement the methods below.
    """

    # MODEL INTEL:
    class Meta(DataModel.Meta):
        store_id = "InSituBehaviour" # Override this so it is a unique string
        concrete = False # Indicates this cannot be instantiated and added in the UI
        
    mixture = property(DataModel.parent.fget, DataModel.parent.fset)

    # REFINEMENT GROUP IMPLEMENTATION:
    @property
    def refine_title(self):
        return "In-situ behaviour"

    @property
    def refine_descriptor_data(self):
        return dict(
            phase_name=self.phase.refine_title,
            component_name="*"
        )

    #: The name of this Behaviour
    name = StringProperty(
        default="New Behaviour", text="Name",
        visible=True, persistent=True, tabular=True
    )
    
    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        my_kwargs = self.pop_kwargs(kwargs,
            *[prop.label for prop in InSituBehaviour.Meta.get_local_persistent_properties()]
        )
        super(InSituBehaviour, self).__init__(*args, **kwargs)
        kwargs = my_kwargs

        with self.data_changed.hold():
            self.name = self.get_kwarg(kwargs, self.name, "name")
            
        pass #end of constructor
            
    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def apply(self, phase):
        assert phase is not None, "Cannot apply on None"
        assert self.is_compatible_with(phase), "`%r` is not compatible with phase `%r`" % (self, phase)
                
    def is_compatible_with(self, phase):
        return False # sub classes need to override this 
               
    pass #end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .edit_mixture_view import EditMixtureView
#from .edit_insitu_mixture_view import EditInSituMixtureView
#from .edit_insitu_behaviour_view import EditInSituBehaviourView
from .add_mixture_view import AddMixtureView 

__all__ = [
    "EditMixtureView",
    #"EditInSituMixtureView",
    "AddMixtureView",
    #"EditInSituBehaviourView"
]


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gdk

import logging
logger = logging.getLogger(__name__)

from mvc.adapters.gtk_support.treemodels.utils import create_treestore_from_directory

from pyxrd.data import settings

from pyxrd.generic.views.combobox_tools import add_combo_text_column
from pyxrd.generic.controllers import DialogController

from pyxrd.probabilities.models import get_Gbounds_for_R, get_Rbounds_for_G

class AddPhaseController(DialogController):
    """ 
        Controller for the add phase dialog
    """

    auto_adapt = False

    def __init__(self, model=None, view=None, parent=None, callback=None):
        super(AddPhaseController, self).__init__(
            model=model, view=view, parent=parent)
        self.callback = callback

    def register_view(self, view):
        self.update_bounds()
        self.generate_combo()

    def register_adapters(self):
        pass # has no intel, or a model!

    def update_R_bounds(self):
        if self.view is not None:
            min_R, max_R, R = get_Rbounds_for_G(
                self.view.get_G(), self.view.get_R())
            self.view["adj_R"].set_upper(max_R)
            self.view["adj_R"].set_lower(min_R)
            self.view["R"].set_value(R)

    def update_G_bounds(self):
        if self.view is not None:
            min_G, max_G, G = get_Gbounds_for_R(
                self.view.get_R(), self.view.get_G())
            self.view["adj_G"].set_upper(max_G)
            self.view["adj_G"].set_lower(min_G)
            self.view["G"].set_value(G)

    def update_bounds(self):
        self.update_G_bounds()
        self.update_R_bounds()

    def generate_combo(self):
        self.reload_combo_model()
        add_combo_text_column(
            self.view.phase_combo_box, text_col=0, sensitive_col=2)

    def reload_combo_model(self):
        cmb_model = create_treestore_from_directory(
            settings.DATA_REG.get_directory_path("DEFAULT_PHASES"))
        self.view.phase_combo_box.set_model(cmb_model)

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_btn_ok_clicked(self, event):
        self.view.hide()
        self.callback(
            self.view.get_phase_type(), self.view.get_G(), self.view.get_R())
        return True

    def on_rdb_toggled(self, widget):
        self.view.update_sensitivities()

    def on_btn_generate_phases_clicked(self, event):
        from pyxrd.scripts.generate_default_phases import run
        def ui_callback(progress):
            self.view["gen_progress_bar"].set_fraction(progress)
            while Gtk.events_pending():
                Gtk.main_iteration()
        self.view["img_repeat"].set_visible(False)
        self.view["gen_spinner"].start()
        self.view["gen_spinner"].set_visible(True)
        self.view["gen_progress_bar"].set_visible(True)
        while Gtk.events_pending():
            Gtk.main_iteration()
        run(ui_callback=ui_callback)
        self.view["gen_progress_bar"].set_visible(False)
        self.view["img_repeat"].set_visible(True)
        self.view["gen_spinner"].stop()
        self.view["gen_spinner"].set_visible(False)
        self.reload_combo_model()
        return True

    def on_r_value_changed(self, *args):
        self.update_G_bounds()
        return True

    def on_g_value_changed(self, *args):
        self.update_R_bounds()
        return True

    def on_keypress(self, widget, event):
        if event.keyval == Gdk.keyval_from_name("Escape"):
            self.view.hide()
            return True
        if event.keyval == Gdk.keyval_from_name("Enter"):
            self.view.hide()
            self.callback(
                self.view.get_phase_type(), self.view.get_G(), self.view.get_R())
            return True

    def on_window_edit_dialog_delete_event(self, event, args=None):
        self.view.hide()
        return True # do not propagate

    pass # end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

class DataObject(object):
    """
        A generic class holding all the information retrieved from a file
        using a BaseParser class.
    """

    # general information
    filename = None

    def __init__(self, *args, **kwargs):
        super(DataObject, self).__init__()
        self.update(**kwargs)

    def update(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    pass # end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import numpy as np

from mvc.models.properties import SetActionMixin, FloatProperty, BoolProperty

from pyxrd.generic.io import storables
from pyxrd.generic.utils import not_none
from pyxrd.generic.models.properties import InheritableMixin
from pyxrd.refinement.refinables.properties import RefinableMixin

from .base_models import _AbstractProbability

def R0_model_generator(pasG):

    class _R0Meta(_AbstractProbability.Meta):
        store_id = "R0G%dModel" % pasG

    class _BaseR0Model(object):
        """
        Probability model for Reichweite = 0
        (g-1) independent variables:
        
        W0 = W0/sum(W0>Wg)
        W1/sum(W1>Wg)
        W2/sum(W2>Wg)
        etc.
        
    
        Pij = Wj
        ∑W = 1
        ∑P = 1
    
        indexes are NOT zero-based in external property names!
        """

        # ------------------------------------------------------------
        #      Initialization and other internals
        # ------------------------------------------------------------
        def __init__(self, *args, **kwargs):
            my_kwargs = self.pop_kwargs(kwargs, *[prop.label for prop in self.Meta.get_local_persistent_properties()])
            super(_BaseR0Model, self).__init__(R=0, *args, **kwargs)

            with self.data_changed.hold():
                if self.G > 1 and "W1" in my_kwargs: # old-style model
                    for i in range(self.G - 1):
                        name = "W%d" % (i + 1)
                        self.mW[i] = not_none(my_kwargs.get(name, None), 0.8)
                        name = "F%d" % (i + 1)
                        setattr(self, name, self.mW[i] / (np.sum(np.diag(self._W)[i:]) or 1.0))
                else:
                    for i in range(self.G - 1):
                        name = "inherit_F%d" % (i + 1)
                        setattr(self, name, my_kwargs.get(name, False))
                        name = "F%d" % (i + 1)
                        setattr(self, name, not_none(my_kwargs.get(name, None), 0.8))

                self.update()

        # ------------------------------------------------------------
        #      Methods & Functions
        # ------------------------------------------------------------
        def update(self):
            with self.monitor_changes():
                if self.G > 1:
                    for i in range(self.G - 1):
                        name = "F%d" % (i + 1)
                        if i > 0:
                            self.mW[i] = getattr(self, name) * (1.0 - np.sum(np.diag(self._W)[0:i]))
                        else:
                            self.mW[i] = getattr(self, name)
                    self.mW[self.G - 1] = 1.0 - np.sum(np.diag(self._W)[:-1])
                else:
                    self.mW[0] = 1.0
                self._P[:] = np.repeat(np.diag(self._W)[np.newaxis, :], self.G, 0)

                self.solve()
                self.validate()

        pass # end of class



    _dict = dict()
    _dict["Meta"] = _R0Meta

    def set_attribute(name, value): # @NoSelf
        """Sets an attribute on the class and the dict"""
        _dict[name] = value
        setattr(_BaseR0Model, name, value)

    set_attribute("G", pasG)

    # PROPERTIES:
    for g in range(pasG - 1):
        label = "F%d" % (g + 1)
        text = "W%(g)d/Sum(W%(g)d+...+W%(G)d)" % {'g':g + 1, 'G':pasG }
        math_text = r"$\large\frac{W_{%(g)d}}{\sum_{i=%(g)d}^{%(G)d} W_i}$" % {'g':g + 1, 'G':pasG }
        inh_flag = "inherit_F%d" % (g + 1)
        inh_from = "parent.based_on.probabilities.F%d" % (g + 1)

        set_attribute(label, FloatProperty(
            default=0.8, text=text, math_text=math_text,
            refinable=True, persistent=True, visible=True,
            minimum=0.0, maximum=1.0,
            inheritable=True, inherit_flag=inh_flag, inherit_from=inh_from,
            is_independent=True, store_private=True,
            set_action_name="update",
            mix_with=(SetActionMixin, RefinableMixin, InheritableMixin)
        ))

        label = "inherit_F%d" % (g + 1)
        text = "Inherit flag for F%d" % (g + 1)
        set_attribute(label, BoolProperty(
            default=False, text=text,
            refinable=False, persistent=True, visible=True,
            set_action_name="update",
            mix_with=(SetActionMixin,)
        ))

    # CREATE TYPE AND REGISTER AS STORABLE:
    cls = type("R0G%dModel" % pasG, (_BaseR0Model, _AbstractProbability), _dict)
    storables.register_decorator(cls)

    return cls

R0G1Model = R0_model_generator(1)
R0G2Model = R0_model_generator(2)
R0G3Model = R0_model_generator(3)
R0G4Model = R0_model_generator(4)
R0G5Model = R0_model_generator(5)
R0G6Model = R0_model_generator(6)

__all__ = [
    "R0G1Model",
    "R0G2Model",
    "R0G3Model",
    "R0G4Model",
    "R0G5Model",
    "R0G6Model"
]



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import numpy as np
from scipy import stats

from .math_tools import smooth

def find_closest(value, array, col=0):
    """
        Find closest value to another value in an array 
    """
    nparray = np.array(list(zip(*array))[col])
    idx = (np.abs(nparray - value)).argmin()
    return array[idx]

def score_minerals(peak_list, minerals):
    """
        Scoring function for mineral peaks
        peak_list: list of tuples containing observed peak position and (abs) intensity
        minerals: dict with the mineral name as key and a peak list as value,
           analogeous to the first argument but with relative intensities
           
        Uses a simple appoach:
        loop over the reference minerals, 
         loop over first 10 strongest peaks for that mineral (starting at 
          stronger reflections towards weaker)
         find a peak in the observed peak data set that matches
         if matching add to the list of matched peaks
         if not matching and this is the strongest reflections, ignore mineral
        after this initial loop we have a list of matched peaks,
        this list is then used to calculate a score for the mineral by looking
        at how well positions and intensities match with the reference and how
        many peaks are actually matched of course. Higher values means a higher
        likelihood this mineral is present.
    """
    max_pos_dev = 0.01 # fraction
    scores = []
    for mineral, abbreviation, mpeaks in minerals:
        tot_score = 0
        p_matches = []
        i_matches = []
        already_matched = []
        mpeaks = sorted(mpeaks, key=lambda peak: peak[0], reverse=True)
        if len(mpeaks) > 15:
            mpeaks = mpeaks[:15]

        for i, (mpos, mint) in enumerate(mpeaks):
            epos, eint = find_closest(mpos, peak_list)
            if abs(epos - mpos) / mpos <= max_pos_dev and not epos in already_matched:
                p_matches.append([mpos, epos])
                i_matches.append([mint, eint])
                already_matched.append(epos)
            elif i == 0:
                break # if strongest peak does not match, ignore mineral

        if len(p_matches) > 3:
            p_matches = np.array(p_matches)
            i_matches = np.array(i_matches)

            i_matches[:, 1] = i_matches[:, 1] / np.max(i_matches[:, 1])

            p_slope, p_intercept, p_r_value, p_value, p_std_err = stats.linregress(p_matches) # @UnusedVariable
            i_slope, i_intercept, i_r_value, p_value, i_std_err = stats.linregress(i_matches) # @UnusedVariable

            p_factor = (p_r_value ** 2) * min(1.0 / (abs(1.0 - p_slope) + 1E-50), 1000.) / 1000.0
            i_factor = (1.0 - min(i_std_err / 0.25, 5.0) / 5.0) * min(1.0 / (abs(1.0 - i_slope) + 1E-50), 1000.) / 1000.0 # * max(1. / (abs(i_intercept) + 1E-50), 100.) / 100.
            tot_score = len(p_matches) * p_factor * i_factor

        if tot_score > 0:
            scores.append((mineral, abbreviation, mpeaks, p_matches, tot_score))

    scores = sorted(scores, key=lambda score: score[-1], reverse=True)
    return scores

def peakdetect(y_axis, x_axis=None, lookahead=500, delta=0):
    """ single run of multi_peakdetect """
    mintabs, maxtabs = multi_peakdetect(y_axis, x_axis, lookahead, [delta])
    return mintabs[0], maxtabs[0]

def multi_peakdetect(y_axis, x_axis=None, lookahead=500, deltas=[0]):
    """
    Converted from/based on a MATLAB script at http://billauer.co.il/peakdet.html
    
    Algorithm for detecting local maximas and minmias in a signal.
    Discovers peaks by searching for values which are surrounded by lower
    or larger values for maximas and minimas respectively
    
    keyword arguments:
    y_axis -- A list containg the signal over which to find peaks
    x_axis -- A x-axis whose values correspond to the 'y_axis' list and is used
        in the return to specify the postion of the peaks. If omitted the index
        of the y_axis is used. (default: None)
    lookahead -- (optional) distance to look ahead from a peak candidate to
        determine if it is the actual peak (default: 500) 
        '(sample / period) / f' where '4 >= f >= 1.25' might be a good value
    deltas -- (optional) this specifies a minimum difference between a peak and
        the following points, before a peak may be considered a peak. Useful
        to hinder the algorithm from picking up false peaks towards to end of
        the signal. To work well delta should be set to 'delta >= RMSnoise * 5'.
        (default: 0)
            Delta function causes a 20% decrease in speed, when omitted
            Correctly used it can double the speed of the algorithm
    
    return -- two lists [maxtab, mintab] containing the positive and negative
        peaks respectively. Each cell of the lists contains a tupple of:
        (position, peak_value) 
        to get the average peak value do 'np.mean(maxtab, 0)[1]' on the results
    """
    rlen = list(range(len(deltas)))
    maxtab = [ [] for i in rlen] # @UnusedVariable
    mintab = [ [] for i in rlen] # @UnusedVariable
    dump = [ [] for i in rlen] # Used to pop the first hit which always if false @UnusedVariable

    length = len(y_axis)
    y_axis = y_axis / np.max(y_axis)
    if x_axis is None:
        x_axis = list(range(length))

    # perform some checks
    if length != len(x_axis):
        raise ValueError("Input vectors y_axis and x_axis must have same length")
    if lookahead < 1:
        raise ValueError("Lookahead must be above '1' in value")

    # needs to be a numpy array
    y_axis = np.asarray(y_axis)

    # Only detect peak if there is 'lookahead' amount of points after it
    for j, delta in enumerate(deltas):

        # maxima and minima candidates are temporarily stored in
        # mx and mn respectively
        mn, mx = np.Inf, -np.Inf

        for index, (x, y) in enumerate(zip(x_axis[:-lookahead], y_axis[:-lookahead])):
            if y > mx:
                mx = y
                mxpos = x
            if y < mn:
                mn = y
                mnpos = x
            ####look for max####
            if y < mx - delta and mx != np.Inf:
                # Maxima peak candidate found
                # look ahead in signal to ensure that this is a peak and not jitter
                if y_axis[index:index + lookahead].max() < mx:
                    maxtab[j].append((mxpos, mx))
                    dump[j].append(True)
                    # set algorithm to only find minima now
                    mx = np.Inf
                    mn = np.Inf

            ####look for min####
            if y > mn + delta and mn != -np.Inf:
                # Minima peak candidate found
                # look ahead in signal to ensure that this is a peak and not jitter
                if y_axis[index:index + lookahead].min() > mn:
                    mintab[j].append((mnpos, mn))
                    dump[j].append(False)
                    # set algorithm to only find maxima now
                    mn = -np.Inf
                    mx = -np.Inf


    # Remove the false hit on the first value of the y_axis
    for j in rlen:
        try:
            if dump[j][0]:
                maxtab[j].pop(0)
            else:
                mintab[j].pop(0)
            # del dump[j]
        except IndexError:
            # no peaks were found, should the function return empty lists?
            pass

    return maxtab, mintab

def peakdetect_zero_crossing(y_axis, x_axis=None, window=49):
    """
    Algorithm for detecting local maximas and minmias in a signal.
    Discovers peaks by dividing the signal into bins and retrieving the
    maximum and minimum value of each the even and odd bins respectively.
    Division into bins is performed by smoothing the curve and finding the
    zero crossings.
    
    Suitable for repeatable sinusoidal signals with some amount of RMS noise
    tolerable. Excecutes faster than 'peakdetect', although this function will
    break if the offset of the signal is too large. It should also be noted
    that the first and last peak will probably not be found, as this algorithm
    only can find peaks between the first and last zero crossing.
    
    keyword arguments:
    y_axis -- A list containg the signal over which to find peaks
    x_axis -- A x-axis whose values correspond to the 'y_axis' list and is used
        in the return to specify the postion of the peaks. If omitted the index
        of the y_axis is used. (default: None)
    window -- the dimension of the smoothing window; should be an odd integer
        (default: 49)
    
    return -- two lists [maxtab, mintab] containing the positive and negative
        peaks respectively. Each cell of the lists contains a tupple of:
        (position, peak_value) 
        to get the average peak value do 'np.mean(maxtab, 0)[1]' on the results
    """
    if x_axis is None:
        x_axis = list(range(len(y_axis)))

    length = len(y_axis)
    if length != len(x_axis):
        raise ValueError('Input vectors y_axis and x_axis must have same length')

    # needs to be a numpy array
    y_axis = np.asarray(y_axis)

    zero_indices = zero_crossings(y_axis, window=window)
    period_lengths = np.diff(zero_indices)

    bins = [y_axis[indice:indice + diff] for indice, diff in
        zip(zero_indices, period_lengths)]

    even_bins = bins[::2]
    odd_bins = bins[1::2]
    # check if even bin contains maxima
    if even_bins[0].max() > abs(even_bins[0].min()):
        hi_peaks = [even.max() for even in even_bins]
        lo_peaks = [odd.min() for odd in odd_bins]
    else:
        hi_peaks = [odd.max() for odd in odd_bins]
        lo_peaks = [even.min() for even in even_bins]


    hi_peaks_x = [x_axis[np.where(y_axis == peak)[0]] for peak in hi_peaks]
    lo_peaks_x = [x_axis[np.where(y_axis == peak)[0]] for peak in lo_peaks]

    maxtab = [(x, y) for x, y in zip(hi_peaks, hi_peaks_x)]
    mintab = [(x, y) for x, y in zip(lo_peaks, lo_peaks_x)]

    return maxtab, mintab

def zero_crossings(y_axis, x_axis=None, window=24):
    """
    Algorithm to find zero crossings. Smoothens the curve and finds the
    zero-crossings by looking for a sign change.
    
    
    keyword arguments:
    y_axis -- A list containg the signal over which to find zero-crossings
    x_axis -- A x-axis whose values correspond to the 'y_axis' list and is used
        in the return to specify the postion of the zero-crossings. If omitted
        then the indice of the y_axis is used. (default: None)
    window -- half of the dimension of the smoothing window;
        (default: 24)
    
    return -- the x_axis value or the indice for each zero-crossing
    """
    # smooth the curve
    length = len(y_axis)
    if x_axis == None:
        x_axis = list(range(length))

    x_axis = np.asarray(x_axis)

    y_axis = smooth(y_axis, window)
    zero_crossings = np.where(np.diff(np.sign(y_axis)))[0]
    times = [x_axis[indice] for indice in zero_crossings]

    # check if zero-crossings are valid
    diff = np.diff(times)
    if diff.std() / diff.mean() > 0.1:
        raise ValueError("smoothing window too small, false zero-crossings found")

    return times



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import os

import numpy as np

from pyxrd.generic.io.utils import get_case_insensitive_glob
from pyxrd.generic.utils import u

from ..base_parser import BaseParser
from .namespace import xrd_parsers
from .xrd_parser_mixin import XRDParserMixin

@xrd_parsers.register_parser()
class UDFParser(XRDParserMixin, BaseParser):
    """
        ASCII Philips *.UDF format
    """

    description = "Philips *.UDF"
    extensions = get_case_insensitive_glob("*.UDF")

    @classmethod
    def _parse_header(cls, filename, fp, data_objects=None, close=False):
        f = fp
        try:
            basename = u(os.path.basename(filename))
        except:
            basename = None
        # Adapt XRDFile list
        data_objects = cls._adapt_data_object_list(data_objects, num_samples=1)

        # Move to the start of the file
        f.seek(0)

        # Go over the header:
        header_dict = {}

        for lineno, line in enumerate(f):
            # Start of data after this line:
            if line.strip() == "RawScan":
                data_start = f.tell()
                break
            else:
                # Break header line into separate parts, and strip trailing whitespace:
                parts = list(map(str.strip, line.split(',')))

                # If length is shorter then three, somethings wrong
                if len(parts) < 3:
                    raise IOError("Header of UDF file is malformed at line %d" % lineno)

                # Handle some of the header's arguments manually, the rest is
                # just passed to the data object as keyword arguments...
                if parts[0] == "SampleIdent":
                    name = parts[1]
                elif parts[0] == "DataAngleRange":
                    twotheta_min = float(parts[1])
                    twotheta_max = float(parts[2])
                elif parts[0] == "ScanStepSize":
                    twotheta_step = float(parts[1])

                # TODO extract other keys and replace with default names
                header_dict[parts[0]] = ','.join(parts[1:-1])

        twotheta_count = int((twotheta_max - twotheta_min) / twotheta_step)

        data_objects[0].update(
            filename=basename,
            name=name,
            twotheta_min=twotheta_min,
            twotheta_max=twotheta_max,
            twotheta_step=twotheta_step,
            twotheta_count=twotheta_count,
            data_start=data_start,
            **header_dict
        )

        if close: f.close()
        return data_objects

    @classmethod
    def _parse_data(cls, filename, fp, data_objects=None, close=False):
        f = fp

        # UDF files are singletons, so no need to iterate over the list,
        # there is only one data object instance:
        if data_objects[0].data == None:
            data_objects[0].data = []

        if f is not None:
            f.seek(data_objects[0].data_start)
            n = 0
            last_value_reached = False
            while n <= data_objects[0].twotheta_count and not last_value_reached:
                parts = list(map(str.strip, f.readline().split(',')))
                for part in parts:
                    # Last value ends with a slash:
                    if part.endswith('/'):
                        part = part[:-1] # remove the ending "/"
                        last_value_reached = True
                    n += 1
                    data_objects[0].data.append([ float(data_objects[0].twotheta_min + data_objects[0].twotheta_step * n), float(part) ])

        data_objects[0].data = np.array(data_objects[0].data)

        if close: f.close()
        return data_objects

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import multiprocessing
from functools import wraps
import traceback, sys

class WrapException(Exception):
    """
        A wrapped exception used by the :meth:`~wrap_exceptions` decorator.
    """
    def __init__(self):
        exc_type, exc_value, exc_tb = sys.exc_info()
        self.exception = exc_value
        self.formatted = ''.join(traceback.format_exception(exc_type, exc_value, exc_tb))
    def __str__(self):
        return '%s\nOriginal traceback:\n%s' % (Exception.__str__(self), self.formatted)

def wrap_exceptions(func):
    """
        Function decorator that allows to provide useable tracebacks when the
        function is called asynchronously and raises an error.
     """
    @wraps(func)
    def exception_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except:
            if multiprocessing.current_process().daemon:
                raise WrapException()
            else:
                raise
    return exception_wrapper


# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

from .labeled_property import LabeledProperty


class UUIDProperty(LabeledProperty):
    """
     A descriptor that handles removing and adding objects to a pool of unique
     objects. Expects the class type of the instance having this property to
     implement a 'remove_object' and 'add_object' method.
    """

    def __set__(self, instance, value):
        type(instance).object_pool.remove_object(instance)
        retval = super(UUIDProperty, self).__set__(instance, value)
        type(instance).object_pool.add_object(instance)
        return retval

    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pkg_resources import resource_filename  # @UnresolvedImport

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk  # @UnresolvedImport

import matplotlib
from matplotlib.figure import Figure
from matplotlib.backends.backend_gtk3cairo import FigureCanvasGTK3Cairo as FigureCanvasGTK
from matplotlib.backends.backend_gtk3 import NavigationToolbar2GTK3 as NavigationToolbar

from pyxrd.generic.views import BaseView

class RefinerView(BaseView):
    """
        A view for the Refiner object
    """
    
    builder = resource_filename(__name__, "glade/refine_results.glade")
    top = "window_refine_results"
    modal = True

    graph_parent = "plot_box"

    def __init__(self, *args, **kwargs):
        BaseView.__init__(self, *args, **kwargs)

        self.graph_parent = self[self.graph_parent]

        self.get_toplevel().set_transient_for(self.parent.get_toplevel())

        self.setup_matplotlib_widget()

    def update_labels(self, initial, best, last):
        self["initial_residual"].set_text("%f" % initial)
        self["best_residual"].set_text("%f" % best)
        self["last_residual"].set_text("%f" % last)

    def setup_matplotlib_widget(self):
        # TODO Create a mixin for this kind of thing!!
        #style = Gtk.Style()
        self.figure = Figure(dpi=72) #, edgecolor=str(style.bg[2]), facecolor=str(style.bg[2]))

        self.figure.subplots_adjust(bottom=0.20)

        self.canvas = FigureCanvasGTK(self.figure)

        box = Gtk.VBox()
        box.pack_start(NavigationToolbar(self.canvas, self.get_top_widget()), False, True, 0)
        box.pack_start(self.canvas, True, True, 0)
        self.graph_parent.add(box)
        self.graph_parent.show_all()

        cdict = {'red': ((0.0, 0.0, 0.0),
                         (0.5, 1.0, 1.0),
                         (1.0, 0.0, 0.0)),
                'green': ((0.0, 0.0, 0.0),
                         (0.5, 1.0, 1.0),
                         (1.0, 0.0, 0.0)),
                'blue': ((0.0, 0.0, 0.0),
                         (0.5, 1.0, 1.0),
                         (1.0, 0.0, 0.0))}
        self.wbw_cmap = matplotlib.colors.LinearSegmentedColormap('WBW', cdict, 256)

    pass # end of class


# coding=UTF-8
# ex:ts=4:sw=4:et:
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#  Copyright (C) 2005 by Roberto Cavada <roboogle@gmail.com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import weakref

class WeakList(list):
    """
        A list subclass referring to its items using weak references if possible 
    """
    
    def __init__(self, items=list()):
        list.__init__(self)
        tuple(map(self.append, items))

    def get_value(self, item):
        try:
            item = item()
        finally:
            return item

    def make_ref(self, item):
        try:
            item = weakref.ref(item, self.remove)
        finally:
            return item

    def __contains__(self, item):
        return list.__contains__(self, self.make_ref(item))

    def __getitem__(self, key):
        if isinstance(key, slice):
            return type(self)(self.get_value(item) for item in list.__getitem__(self, key))
        return self.get_value(list.__getitem__(self, key))

    def __getslice__(self, i, j):
        return self.__getitem__(slice(i, j))

    def __setitem__(self, key, item):
        return list.__setitem__(self, key, self.make_ref(item))

    def __iter__(self):
        return iter([self[key] for key in range(len(self))])

    def append(self, item):
        list.append(self, self.make_ref(item))

    def remove(self, item):
        item = self.make_ref(item)
        while list.__contains__(self, item):
            list.remove(self, item)

    def index(self, item):
        return list.index(self, self.make_ref(item))

    def pop(self, item):
        return list.pop(self, self.make_ref(item))

    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pkg_resources import resource_filename # @UnresolvedImport

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from pyxrd.generic.views import DialogView

class AddMixtureView(DialogView):
    title = "Add Mixture"
    subview_builder = resource_filename(__name__, "glade/add_mixture.glade")
    subview_toplevel = "add_mixture_container"

    active_type = "mixture" # | insitu

    def __init__(self, type_dict, *args, **kwargs):
        DialogView.__init__(self, *args, **kwargs)
        self.type_dict = type_dict
        self.active_type=list(type_dict.keys())[0]
        self.create_radios()

    def create_radios(self):
        box = self["add_mixture_box"]
        box.clear()
        group = None
        self.radios = []
        for mixture_type, label in self.type_dict.items():
            radio = Gtk.RadioButton.new(group, label)
            radio.mixture_type = mixture_type
            group = radio if group is None else group 
            self.radios.append(radio)
            radio.connect('toggled', self.on_rdb_toggled)
            box.pack_start(radio, False, False, 2)

    def get_mixture_type(self):
        return self.active_type

    def on_rdb_toggled(self):
        for radio in self.radios:
            if radio.get_active():
                self.active_type = radio.mixture_type
    pass # end of class